"""
Speaker verification loss

Authors:
  * Haibin Wu 2022
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = [
    "softmax",
    "amsoftmax",
]


class softmax(nn.Module):
    """
    The standard softmax loss in an unified interface for all speaker-related softmax losses
    """

    def __init__(self, input_size: int, output_size: int):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size

        self.fc = nn.Linear(input_size, output_size)
        self.criertion = nn.CrossEntropyLoss()

    @property
    def input_size(self):
        return self._indim

    @property
    def output_size(self):
        return self._outdim

    def forward(self, x: torch.Tensor, label: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, input_size)
            label (torch.LongTensor): (batch_size, )

        Returns:
            loss (torch.float)
            logit (torch.Tensor): (batch_size, )
        """

        assert x.size()[0] == label.size()[0]
        assert x.size()[1] == self.input_size

        x = F.normalize(x, dim=1)
        x = self.fc(x)
        loss = self.criertion(x, label)

        return loss, x


class amsoftmax(nn.Module):
    """
    AMSoftmax

    Args:
        input_size (int): The input feature size
        output_size (int): The output feature size
        margin (float): Hyperparameter denotes the margin to the decision boundry
        scale (float): Hyperparameter that scales the cosine value
    """

    def __init__(
        self, input_size: int, output_size: int, margin: float = 0.2, scale: float = 30
    ):
        super().__init__()
        self._indim = input_size
        self._outdim = output_size
        self.margin = margin
        self.scale = scale

        self.W = torch.nn.Parameter(
            torch.randn(input_size, output_size), requires_grad=True
        )
        self.ce = nn.CrossEntropyLoss()
        nn.init.xavier_normal_(self.W, gain=1)

    @property
    def input_size(self):
        return self._indim

    @property
    def output_size(self):
        return self._outdim

    def forward(self, x: torch.Tensor, label: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, input_size)
            label (torch.LongTensor): (batch_size, )

        Returns:
            loss (torch.float)
            logit (torch.Tensor): (batch_size, )
        """

        assert x.size()[0] == label.size()[0]
        assert x.size()[1] == self.input_size

        x_norm = torch.norm(x, p=2, dim=1, keepdim=True).clamp(min=1e-12)
        x_norm = torch.div(x, x_norm)
        w_norm = torch.norm(self.W, p=2, dim=0, keepdim=True).clamp(min=1e-12)
        w_norm = torch.div(self.W, w_norm)
        costh = torch.mm(x_norm, w_norm)
        label_view = label.view(-1, 1)
        if label_view.is_cuda:
            label_view = label_view.cpu()
        delt_costh = torch.zeros(costh.size()).scatter_(1, label_view, self.margin)
        if x.is_cuda:
            delt_costh = delt_costh.cuda()
        costh_m = costh - delt_costh
        costh_m_s = self.scale * costh_m
        loss = self.ce(costh_m_s, label)

        return loss, costh_m_s



"""
Dump feature Task

Authors
  * Yist Y. Lin 2021
  * Leo 2022
"""

from pathlib import Path

import torch
import torch.nn as nn

from .base import Task

__all__ = ["DumpFeature"]


class DumpFeature(Task):
    def __init__(self, model: nn.Module, dump_feat_dir: str = "feat") -> None:
        super().__init__()
        self.model = model
        self.dump_feat_dir = dump_feat_dir

    @torch.no_grad()
    def forward(self, split: str, x, x_len, unique_name, _dump_dir: str):
        self.model.eval()
        feats, feats_len = self.model(x, x_len).slice(2)
        feats = torch.stack(
            feats, dim=1
        )  # (batch_size, num_layer, seqlen, hidden_size)

        feat_dir = Path(_dump_dir) / self.dump_feat_dir
        for feat, feat_len, name in zip(feats, feats_len, unique_name):
            feat = feat[:, :feat_len, :]
            feat = feat.detach().cpu()
            torch.save(feat, str(feat_dir / f"{name}.pt"))

        pseudo_loss = torch.zeros(1, requires_grad=True)
        return pseudo_loss, {}

    def reduction(self, split: str, batch_results: list, _dump_dir: str):
        return {}



from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass

from .base import AugmentedDynamicItemDataset, DataPipe


def _count_frames(data_len, size, step):
    return int((data_len - size + step) / step)


def _gen_frame_indices(
    data_length,
    size=2000,
    step=2000,
    use_last_samples=True,
):
    i = -1
    for i in range(_count_frames(data_length, size, step)):
        yield i * step, i * step + size
    if use_last_samples and i * step + size < data_length:
        if data_length - (i + 1) * step > 0:
            yield (i + 1) * step, data_length


@dataclass
class UnfoldChunkByFrame(DataPipe):
    """
    Given a dataset with items containing `start_sec_name` and `end_sec_name`.
    For each item, produce `((end_sec_name - start_sec_name) * sample_rate / feat_frame_shift) / chunk_frames`
    items with the smaller durations between `min_chunk_frames` and `max_chunk_frames`

    Args:
        sample_rate (int): The sample_rate of the audio.
        feat_frame_shift (int): The target feature's frame_shift
        min_chunk_frames (int): The min frames for a chunk
        max_chunk_frames (int): The max frames for a chunk
        step_frames (int): The step frames for each sliding window
        use_last_samples (bool): whether to drop the last samples of an utterance which cannot form an window
        start_sec_name (str): The key name for the starting second of the audio
        end_sec_name (str): The key name for the ending second of the audio
    """

    sample_rate: int = 16000
    feat_frame_shift: int = 160

    min_chunk_frames: int = 2000
    max_chunk_frames: int = 2000
    step_frames: int = 2000
    use_last_samples: bool = True

    start_sec_name: str = "start_sec"
    end_sec_name: str = "end_sec"

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        unfolded_items = OrderedDict()
        for item in dataset:
            key = item.pop("id")

            # TODO: Check whether int v.s. round affect the final result
            # Original SUPERB uses int, however it makes more sense to use round
            data_len = int(
                (item[self.end_sec_name] - item[self.start_sec_name])
                * self.sample_rate
                / self.feat_frame_shift
            )
            for unfold_index, (start, end) in enumerate(
                _gen_frame_indices(data_len, self.min_chunk_frames, self.step_frames)
            ):
                start_sec = (
                    item[self.start_sec_name]
                    + start * self.feat_frame_shift / self.sample_rate
                )
                end_sec = (
                    item[self.start_sec_name]
                    + end * self.feat_frame_shift / self.sample_rate
                )
                dur_sec = end_sec - start_sec
                utt_id = f"{key}_start-{start_sec}_end-{end_sec}_dur-{dur_sec}"
                subitem = deepcopy(item)
                subitem["unchunked_id"] = key
                subitem["chunk_index"] = unfold_index
                subitem[self.start_sec_name] = start_sec
                subitem[self.end_sec_name] = end_sec
                unfolded_items[utt_id] = subitem
        new_dataset = AugmentedDynamicItemDataset(unfolded_items)
        new_dataset.add_tools(dataset.all_tools(False))
        return new_dataset


@dataclass
class UnfoldChunkBySec(DataPipe):
    use_last_samples: bool = True
    min_chunk_secs: float = 2.5
    max_chunk_secs: float = 2.5
    step_secs: int = 2.5

    start_sec_name: str = "start_sec"
    end_sec_name: str = "end_sec"

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        unfolded_items = OrderedDict()
        for item in dataset:
            key = item.pop("id")
            for unfold_index, (start, end) in enumerate(
                _gen_frame_indices(
                    item[self.end_sec_name] - item[self.start_sec_name],
                    self.min_chunk_secs,
                    self.step_secs,
                )
            ):
                start_sec = item[self.start_sec_name] + start
                end_sec = item[self.start_sec_name] + end
                dur_sec = end_sec - start_sec
                utt_id = f"{key}_start-{start_sec}_end-{end_sec}_dur-{dur_sec}"
                subitem = deepcopy(item)
                subitem["unchunked_id"] = key
                subitem["chunk_index"] = unfold_index
                subitem[self.start_sec_name] = start_sec
                subitem[self.end_sec_name] = end_sec
                unfolded_items[utt_id] = subitem
        new_dataset = AugmentedDynamicItemDataset(unfolded_items)
        new_dataset.add_tools(dataset.all_tools(False))
        return new_dataset



"""
The backbone run procedure of ASV tasks

Authors
  * Haibin Wu 2022
  * Leo 2022
"""

import logging
import os
import pickle
import shutil
from pathlib import Path
from typing import List

import pandas as pd
import yaml

from s3prl.problem.base import Problem
from s3prl.task.speaker_verification_task import SpeakerVerification

logger = logging.getLogger(__name__)

__all__ = ["ASV"]


class ASV(Problem):
    def run(
        self,
        target_dir: str,
        cache_dir: str,
        remove_all_cache: bool = False,
        start: int = 0,
        stop: int = None,
        num_workers: int = 6,
        eval_batch: int = -1,
        device: str = "cuda",
        world_size: int = 1,
        rank: int = 0,
        test_ckpt_dir: str = None,
        test_ckpt_steps: List[int] = None,
        prepare_data: dict = None,
        build_encoder: dict = None,
        build_dataset: dict = None,
        build_batch_sampler: dict = None,
        build_collate_fn: dict = None,
        build_upstream: dict = None,
        build_featurizer: dict = None,
        build_downstream: dict = None,
        build_model: dict = None,
        build_task: dict = None,
        build_optimizer: dict = None,
        build_scheduler: dict = None,
        save_model: dict = None,
        save_task: dict = None,
        train: dict = None,
        evaluate: dict = None,
    ):
        """
        ========  ====================
        stage     description
        ========  ====================
        0         Parse the corpus and save the metadata file (waveform path, label...)
        1         Build the encoder for encoding the speaker labels
        2         Train the model
        3         Evaluate the model on multiple test sets, multiple checkpoints will be evaluated for each test set (See :code:`test_ckpt_steps`)
        4         Report the best result find on each test set
        ========  ====================

        Args:
            target_dir (str):
                The directory that stores the script result.
            cache_dir (str):
                The directory that caches the processed data.
                Default: /home/user/.cache/s3prl/data
            remove_all_cache (bool):
                Whether to remove all the cache stored under `cache_dir`.
                Default: False
            start (int):
                The starting stage of the problem script.
                Default: 0
            stop (int):
                The stoping stage of the problem script, set `None` to reach the final stage.
                Default: None
            num_workers (int): num_workers for all the torch DataLoder
            eval_batch (int):
                During evaluation (valid or test), limit the number of batch.
                This is helpful for the fast development to check everything won't crash.
                If is -1, disable this feature and evaluate the entire epoch.
                Default: -1
            device (str):
                The device type for all torch-related operation: "cpu" or "cuda"
                Default: "cuda"
            world_size (int):
                How many processes are running this script simultaneously (in parallel).
                Usually this is just 1, however if you are runnig distributed training,
                this should be > 1.
                Default: 1
            rank (int):
                When distributed training, world_size > 1. Take :code:`world_size == 8` for
                example, this means 8 processes (8 GPUs) are runing in parallel. The script
                needs to know which process among 8 processes it is. In this case, :code:`rank`
                can range from 0~7. All the 8 processes have the same :code:`world_size` but
                different :code:`rank` (process id).
            test_ckpt_dir (str):
                Specify the checkpoint path for testing. If not, use checkpoints specified by
                :code:`test_ckpts_steps`.
            test_ckpt_steps (List[int]):
                After training, multiple steps of checkpoints are saved. This option specifies
                which checkpoints (multiple) will be used for evaluation.
            **kwds:
                The other arguments like :code:`prepare_data` and :code:`build_model` are
                method specific-arguments for methods like :obj:`prepare_data` and
                :obj:`build_model`, and will not be used in the core :obj:`run` logic.
                See the specific method documentation for their supported arguments and
                meaning
        """
        yaml_path = Path(target_dir) / "configs" / f"{self._get_time_tag()}.yaml"
        yaml_path.parent.mkdir(exist_ok=True, parents=True)
        with yaml_path.open("w") as f:
            yaml.safe_dump(self._get_current_arguments(), f)

        cache_dir: str = cache_dir or Path.home() / ".cache" / "s3prl" / "data"
        prepare_data: dict = prepare_data or {}
        build_encoder: dict = build_encoder or {}
        build_dataset: dict = build_dataset or {}
        build_batch_sampler: dict = build_batch_sampler or {}
        build_collate_fn: dict = build_collate_fn or {}
        build_upstream: dict = build_upstream or {}
        build_featurizer: dict = build_featurizer or {}
        build_downstream: dict = build_downstream or {}
        build_model: dict = build_model or {}
        build_task: dict = build_task or {}
        build_optimizer: dict = build_optimizer or {}
        build_scheduler: dict = build_scheduler or {}
        save_model: dict = save_model or {}
        save_task: dict = save_task or {}
        train: dict = train or {}
        evaluate = evaluate or {}

        target_dir: Path = Path(target_dir)
        target_dir.mkdir(exist_ok=True, parents=True)

        cache_dir = Path(cache_dir)
        cache_dir.mkdir(exist_ok=True, parents=True)
        if remove_all_cache:
            shutil.rmtree(cache_dir)

        stage_id = 0
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: prepare data")
            train_csv, test_csvs = self.prepare_data(
                prepare_data, target_dir, cache_dir, get_path_only=False
            )

        train_csv, test_csvs = self.prepare_data(
            prepare_data, target_dir, cache_dir, get_path_only=True
        )

        def check_fn():
            assert Path(train_csv).is_file()
            for test_csv in test_csvs:
                assert Path(test_csv).is_file()

        self._stage_check(stage_id, stop, check_fn)

        stage_id = 1
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: build encoder")
            encoder_path = self.build_encoder(
                build_encoder,
                target_dir,
                cache_dir,
                train_csv,
                test_csvs,
                get_path_only=False,
            )

        encoder_path = self.build_encoder(
            build_encoder,
            target_dir,
            cache_dir,
            train_csv,
            test_csvs,
            get_path_only=True,
        )

        def check_fn():
            assert Path(encoder_path).is_file()

        self._stage_check(stage_id, stop, check_fn)

        stage_id = 2
        train_dir = target_dir / "train"
        if start <= stage_id:
            logger.info(f"Stage {stage_id}: Train Model")
            train_ds, train_bs = self._build_dataset_and_sampler(
                target_dir,
                cache_dir,
                "train",
                train_csv,
                encoder_path,
                build_dataset,
                build_batch_sampler,
            )

            with Path(encoder_path).open("rb") as f:
                encoder = pickle.load(f)

            build_model_all_args = dict(
                build_model=build_model,
                model_output_size=len(encoder),
                build_upstream=build_upstream,
                build_featurizer=build_featurizer,
                build_downstream=build_downstream,
            )
            build_task_all_args_except_model = dict(
                build_task=build_task,
                encoder=encoder,
            )

            self.train(
                train,
                train_dir,
                build_model_all_args,
                build_task_all_args_except_model,
                save_model,
                save_task,
                build_optimizer,
                build_scheduler,
                evaluate,
                train_ds,
                train_bs,
                self.build_collate_fn(build_collate_fn, "train"),
                None,
                None,
                None,
                device=device,
                eval_batch=eval_batch,
                num_workers=num_workers,
                world_size=world_size,
                rank=rank,
            )

        test_ckpt_dirs = []
        if test_ckpt_dir is not None:
            test_ckpt_dirs.append(test_ckpt_dir)
        if test_ckpt_steps is None:
            train_ckpts = [
                train_dir / name
                for name in os.listdir(train_dir)
                if name.startswith("step_")
            ]
            test_ckpt_dirs.extend(train_ckpts)
        else:
            test_ckpt_dirs.extend(
                [train_dir / f"step_{step}" for step in test_ckpt_steps]
            )

        def check_fn():
            for ckpt_dir in test_ckpt_dirs:
                assert Path(ckpt_dir).is_dir(), ckpt_dir

        self._stage_check(stage_id, stop, check_fn)

        stage_id = 3
        if start <= stage_id:
            for test_idx, test_csv in enumerate(test_csvs):
                for ckpt_idx, ckpt_dir in enumerate(test_ckpt_dirs):
                    test_name = Path(test_csv).stem
                    test_dir: Path = (
                        target_dir
                        / "evaluate"
                        / test_name
                        / ckpt_dir.relative_to(train_dir).as_posix().replace("/", "-")
                    )
                    test_dir.mkdir(exist_ok=True, parents=True)

                    logger.info(
                        f"Stage {stage_id}.{test_idx}.{ckpt_idx}: Test on {test_csv} with model {ckpt_dir}"
                    )
                    test_ds, test_bs = self._build_dataset_and_sampler(
                        target_dir,
                        cache_dir,
                        "test",
                        test_csv,
                        encoder_path,
                        build_dataset,
                        build_batch_sampler,
                    )

                    csv = pd.read_csv(test_csv)
                    test_trials = []
                    for rowid, row in csv.iterrows():
                        test_trials.append(
                            (int(row["label"]), str(row["id1"]), str(row["id2"]))
                        )

                    overrides = dict(test_trials=test_trials)
                    _, task = self.load_model_and_task(ckpt_dir, overrides)
                    logs = self.evaluate(
                        evaluate,
                        "test",
                        task,
                        test_ds,
                        test_bs,
                        self.build_collate_fn(build_collate_fn, "test"),
                        eval_batch,
                        test_dir,
                        device,
                        num_workers,
                    )
                    test_metrics = {name: float(value) for name, value in logs.items()}
                    logger.info(f"test metrics: {test_metrics}")

                    assert "EER" in test_metrics
                    with (test_dir / f"result.yaml").open("w") as f:
                        yaml.safe_dump(test_metrics, f)

        self._stage_check(stage_id, stop, lambda: True)

        stage_id = 4
        if start <= stage_id:
            for test_idx, test_csv in enumerate(test_csvs):
                test_name = Path(test_csv).stem
                logger.info(f"Report results on {test_name}")

                eer_ckpts = []
                for ckpt_dir in os.listdir(target_dir / "evaluate" / test_name):
                    result_yaml: Path = (
                        target_dir / "evaluate" / test_name / ckpt_dir / "result.yaml"
                    )
                    if result_yaml.is_file():
                        with open(result_yaml) as f:
                            eer_ckpts.append(
                                (
                                    float(yaml.load(f, Loader=yaml.FullLoader)["EER"]),
                                    str(result_yaml.parent),
                                )
                            )

                logger.info(f"All EERs on {test_name}:")
                for eer, ckpt in eer_ckpts:
                    logger.info(f"ckpt_dir: {ckpt}, eer: {eer}")

                eer_ckpts.sort(key=lambda x: x[0])
                best_eer, best_ckpt_dir = eer_ckpts[0]

                logger.info(
                    f"Best EER on {test_name} is from {best_ckpt_dir}: {best_eer}"
                )

                with (target_dir / "evaluate" / test_name / "best_result.yaml").open(
                    "w"
                ) as f:
                    yaml.safe_dump({"EER": best_eer}, f)

    def _build_dataset_and_sampler(
        self,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
        build_dataset: dict,
        build_batch_sampler: dict,
    ):
        logger.info(f"Build {mode} dataset")
        dataset = self.build_dataset(
            build_dataset,
            target_dir,
            cache_dir,
            mode,
            data_csv,
            encoder_path,
        )
        logger.info(f"Build {mode} batch sampler")
        batch_sampler = self.build_batch_sampler(
            build_batch_sampler,
            target_dir,
            cache_dir,
            mode,
            data_csv,
            dataset,
        )
        return dataset, batch_sampler

    def build_task(
        self,
        build_task: dict,
        model,
        encoder,
        test_trials=None,
    ):
        """
        Build the task, which defines the logics for every train/valid/test forward step for the :code:`model`,
        and the logics for how to reduce all the batch results from multiple train/valid/test steps into metrics

        By default build :obj:`SpeakerVerification`

        Args:
            build_task (dict): same in :obj:`default_config`, no argument supported for now
            model (torch.nn.Module): the model built by :obj:`build_model`
            encoder: the encoder built by :obj:`build_encoder`
            test_trials (List[Tuple[int, str, str]]): each tuple in the list consists of
                :code:`(label, enroll_utt_id, test_utt_id)`. label is either 0 or 1

        Returns:
            Task
        """
        task = SpeakerVerification(model, encoder, test_trials, **build_task)
        return task



import torch
import torch.nn as nn
import torchaudio
from torch.nn.functional import conv1d, conv2d

sz_float = 4  # size of a float
epsilon = 10e-8  # fudge factor for normalization


class AugmentMelSTFT(nn.Module):
    def __init__(
        self,
        n_mels=128,
        sr=32000,
        win_length=800,
        hopsize=320,
        n_fft=1024,
        freqm=48,
        timem=192,
        htk=False,
        fmin=0.0,
        fmax=None,
        norm=1,
        fmin_aug_range=1,
        fmax_aug_range=1000,
    ):
        torch.nn.Module.__init__(self)
        # adapted from: https://github.com/CPJKU/kagglebirds2020/commit/70f8308b39011b09d41eb0f4ace5aa7d2b0e806e
        # Similar config to the spectrograms used in AST: https://github.com/YuanGongND/ast

        self.win_length = win_length
        self.n_mels = n_mels
        self.n_fft = n_fft
        self.sr = sr
        self.htk = htk
        self.fmin = fmin

        if fmax is None:
            fmax = sr // 2 - fmax_aug_range // 2

        self.fmax = fmax
        self.norm = norm
        self.hopsize = hopsize
        self.register_buffer(
            "window", torch.hann_window(win_length, periodic=False), persistent=False
        )
        assert (
            fmin_aug_range >= 1
        ), f"fmin_aug_range={fmin_aug_range} should be >=1; 1 means no augmentation"
        assert (
            fmin_aug_range >= 1
        ), f"fmax_aug_range={fmax_aug_range} should be >=1; 1 means no augmentation"
        self.fmin_aug_range = fmin_aug_range
        self.fmax_aug_range = fmax_aug_range

        self.register_buffer(
            "preemphasis_coefficient", torch.as_tensor([[[-0.97, 1]]]), persistent=False
        )
        if freqm == 0:
            self.freqm = torch.nn.Identity()
        else:
            self.freqm = torchaudio.transforms.FrequencyMasking(freqm, iid_masks=True)
        if timem == 0:
            self.timem = torch.nn.Identity()
        else:
            self.timem = torchaudio.transforms.TimeMasking(timem, iid_masks=True)

    def forward(self, x):
        x = nn.functional.conv1d(x.unsqueeze(1), self.preemphasis_coefficient).squeeze(
            1
        )
        x = torch.stft(
            x,
            self.n_fft,
            hop_length=self.hopsize,
            win_length=self.win_length,
            center=True,
            normalized=False,
            window=self.window,
            return_complex=True,
        )
        x = torch.view_as_real(x)
        x = (x**2).sum(dim=-1)  # power mag
        fmin = self.fmin + torch.randint(self.fmin_aug_range, (1,)).item()
        fmax = (
            self.fmax
            + self.fmax_aug_range // 2
            - torch.randint(self.fmax_aug_range, (1,)).item()
        )
        # don't augment eval data
        if not self.training:
            fmin = self.fmin
            fmax = self.fmax

        mel_basis, _ = torchaudio.compliance.kaldi.get_mel_banks(
            self.n_mels,
            self.n_fft,
            self.sr,
            fmin,
            fmax,
            vtln_low=100.0,
            vtln_high=-500.0,
            vtln_warp_factor=1.0,
        )
        mel_basis = torch.as_tensor(
            torch.nn.functional.pad(mel_basis, (0, 1), mode="constant", value=0),
            device=x.device,
        )
        with torch.cuda.amp.autocast(enabled=False):
            melspec = torch.matmul(mel_basis, x)

        melspec = (melspec + 0.00001).log()

        if self.training:
            melspec = self.freqm(melspec)
            melspec = self.timem(melspec)

        melspec = (melspec + 4.5) / 5.0  # fast normalization

        return melspec

    def extra_repr(self):
        return "winsize={}, hopsize={}".format(self.win_length, self.hopsize)



# -*- coding: utf-8 -*-
# @Time    : 6/10/21 5:04 PM
# @Author  : Yuan Gong
# @Affiliation  : Massachusetts Institute of Technology
# @Email   : yuangong@mit.edu
# @File    : ast_models.py

import logging

import timm
import torch
import torch.nn as nn
from timm.models.layers import to_2tuple, trunc_normal_
from torch.cuda.amp import autocast

logger = logging.getLogger(__name__)

NUM_VERTICAL_PATCH = 12


# override the timm package to relax the input shape constraint.
class PatchEmbed(nn.Module):
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()

        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )

    def forward(self, x):
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


class ASTModel(nn.Module):
    """
    The AST model.
    :param label_dim: the label dimension, i.e., the number of total classes, it is 527 for AudioSet, 50 for ESC-50, and 35 for speechcommands v2-35
    :param fstride: the stride of patch spliting on the frequency dimension, for 16*16 patchs, fstride=16 means no overlap, fstride=10 means overlap of 6
    :param tstride: the stride of patch spliting on the time dimension, for 16*16 patchs, tstride=16 means no overlap, tstride=10 means overlap of 6
    :param input_fdim: the number of frequency bins of the input spectrogram
    :param input_tdim: the number of time frames of the input spectrogram
    :param imagenet_pretrain: if use ImageNet pretrained model
    :param audioset_pretrain: if use full AudioSet and ImageNet pretrained model
    :param model_size: the model size of AST, should be in [tiny224, small224, base224, base384], base224 and base 384 are same model, but are trained differently during ImageNet pretraining.
    """

    def __init__(
        self,
        label_dim=527,
        fstride=10,
        tstride=10,
        input_fdim=128,
        input_tdim=1024,
        imagenet_pretrain=True,
        audioset_pretrain=False,
        model_size="base384",
        verbose=True,
        pretrained_ckpt=None,
    ):
        super(ASTModel, self).__init__()
        assert (
            timm.__version__ == "0.4.5"
        ), "Please use timm == 0.4.5, the code might not be compatible with newer versions."

        if verbose == True:
            logger.info("---------------AST Model Summary---------------")
            logger.info(
                "ImageNet pretraining: {:s}, AudioSet pretraining: {:s}".format(
                    str(imagenet_pretrain), str(audioset_pretrain)
                )
            )
        # override timm input shape restriction
        timm.models.vision_transformer.PatchEmbed = PatchEmbed

        # if AudioSet pretraining is not used (but ImageNet pretraining may still apply)
        if audioset_pretrain == False:
            if model_size == "tiny224":
                self.v = timm.create_model(
                    "vit_deit_tiny_distilled_patch16_224", pretrained=imagenet_pretrain
                )
            elif model_size == "small224":
                self.v = timm.create_model(
                    "vit_deit_small_distilled_patch16_224", pretrained=imagenet_pretrain
                )
            elif model_size == "base224":
                self.v = timm.create_model(
                    "vit_deit_base_distilled_patch16_224", pretrained=imagenet_pretrain
                )
            elif model_size == "base384":
                self.v = timm.create_model(
                    "vit_deit_base_distilled_patch16_384", pretrained=imagenet_pretrain
                )
            else:
                raise Exception(
                    "Model size must be one of tiny224, small224, base224, base384."
                )
            self.original_num_patches = self.v.patch_embed.num_patches
            self.oringal_hw = int(self.original_num_patches**0.5)
            self.original_embedding_dim = self.v.pos_embed.shape[2]
            self.mlp_head = nn.Sequential(
                nn.LayerNorm(self.original_embedding_dim),
                nn.Linear(self.original_embedding_dim, label_dim),
            )

            # automatcially get the intermediate shape
            f_dim, t_dim = self.get_shape(fstride, tstride, input_fdim, input_tdim)
            num_patches = f_dim * t_dim
            self.v.patch_embed.num_patches = num_patches
            if verbose == True:
                logger.info(
                    "frequncey stride={:d}, time stride={:d}".format(fstride, tstride)
                )
                logger.info("number of patches={:d}".format(num_patches))

            # the linear projection layer
            new_proj = torch.nn.Conv2d(
                1,
                self.original_embedding_dim,
                kernel_size=(16, 16),
                stride=(fstride, tstride),
            )
            if imagenet_pretrain == True:
                new_proj.weight = torch.nn.Parameter(
                    torch.sum(self.v.patch_embed.proj.weight, dim=1).unsqueeze(1)
                )
                new_proj.bias = self.v.patch_embed.proj.bias
            self.v.patch_embed.proj = new_proj

            # the positional embedding
            if imagenet_pretrain == True:
                # get the positional embedding from deit model, skip the first two tokens (cls token and distillation token), reshape it to original 2D shape (24*24).
                new_pos_embed = (
                    self.v.pos_embed[:, 2:, :]
                    .detach()
                    .reshape(1, self.original_num_patches, self.original_embedding_dim)
                    .transpose(1, 2)
                    .reshape(
                        1, self.original_embedding_dim, self.oringal_hw, self.oringal_hw
                    )
                )
                # cut (from middle) or interpolate the second dimension of the positional embedding
                if t_dim <= self.oringal_hw:
                    new_pos_embed = new_pos_embed[
                        :,
                        :,
                        :,
                        int(self.oringal_hw / 2)
                        - int(t_dim / 2) : int(self.oringal_hw / 2)
                        - int(t_dim / 2)
                        + t_dim,
                    ]
                else:
                    new_pos_embed = torch.nn.functional.interpolate(
                        new_pos_embed, size=(self.oringal_hw, t_dim), mode="bilinear"
                    )
                # cut (from middle) or interpolate the first dimension of the positional embedding
                if f_dim <= self.oringal_hw:
                    new_pos_embed = new_pos_embed[
                        :,
                        :,
                        int(self.oringal_hw / 2)
                        - int(f_dim / 2) : int(self.oringal_hw / 2)
                        - int(f_dim / 2)
                        + f_dim,
                        :,
                    ]
                else:
                    new_pos_embed = torch.nn.functional.interpolate(
                        new_pos_embed, size=(f_dim, t_dim), mode="bilinear"
                    )
                # flatten the positional embedding
                new_pos_embed = new_pos_embed.reshape(
                    1, self.original_embedding_dim, num_patches
                ).transpose(1, 2)
                # concatenate the above positional embedding with the cls token and distillation token of the deit model.
                self.v.pos_embed = nn.Parameter(
                    torch.cat(
                        [self.v.pos_embed[:, :2, :].detach(), new_pos_embed], dim=1
                    )
                )
            else:
                # if not use imagenet pretrained model, just randomly initialize a learnable positional embedding
                # TODO can use sinusoidal positional embedding instead
                new_pos_embed = nn.Parameter(
                    torch.zeros(
                        1,
                        self.v.patch_embed.num_patches + 2,
                        self.original_embedding_dim,
                    )
                )
                self.v.pos_embed = new_pos_embed
                trunc_normal_(self.v.pos_embed, std=0.02)

        # now load a model that is pretrained on both ImageNet and AudioSet
        elif audioset_pretrain == True:
            if audioset_pretrain == True and imagenet_pretrain == False:
                raise ValueError(
                    "currently model pretrained on only audioset is not supported, please set imagenet_pretrain = True to use audioset pretrained model."
                )
            if model_size != "base384":
                raise ValueError(
                    "currently only has base384 AudioSet pretrained model."
                )
            sd = torch.load(pretrained_ckpt, map_location="cpu")
            audio_model = ASTModel(
                label_dim=527,
                fstride=10,
                tstride=10,
                input_fdim=128,
                input_tdim=1024,
                imagenet_pretrain=False,
                audioset_pretrain=False,
                model_size="base384",
                verbose=False,
            )
            audio_model = torch.nn.DataParallel(audio_model)
            audio_model.load_state_dict(sd, strict=False)
            self.v = audio_model.module.v
            self.original_embedding_dim = self.v.pos_embed.shape[2]
            self.mlp_head = nn.Sequential(
                nn.LayerNorm(self.original_embedding_dim),
                nn.Linear(self.original_embedding_dim, label_dim),
            )

            f_dim, t_dim = self.get_shape(fstride, tstride, input_fdim, input_tdim)
            num_patches = f_dim * t_dim
            self.v.patch_embed.num_patches = num_patches
            if verbose == True:
                logger.info(
                    "frequncey stride={:d}, time stride={:d}".format(fstride, tstride)
                )
                logger.info("number of patches={:d}".format(num_patches))

            new_pos_embed = (
                self.v.pos_embed[:, 2:, :]
                .detach()
                .reshape(1, 1212, 768)
                .transpose(1, 2)
                .reshape(1, 768, 12, 101)
            )
            # if the input sequence length is larger than the original audioset (10s), then cut the positional embedding
            if t_dim < 101:
                new_pos_embed = new_pos_embed[
                    :, :, :, 50 - int(t_dim / 2) : 50 - int(t_dim / 2) + t_dim
                ]
            # otherwise interpolate
            else:
                new_pos_embed = torch.nn.functional.interpolate(
                    new_pos_embed, size=(12, t_dim), mode="bilinear"
                )
            if f_dim < 12:
                new_pos_embed = new_pos_embed[
                    :, :, 6 - int(f_dim / 2) : 6 - int(f_dim / 2) + f_dim, :
                ]
            # otherwise interpolate
            elif f_dim > 12:
                new_pos_embed = torch.nn.functional.interpolate(
                    new_pos_embed, size=(f_dim, t_dim), mode="bilinear"
                )
            new_pos_embed = new_pos_embed.reshape(1, 768, num_patches).transpose(1, 2)
            self.v.pos_embed = nn.Parameter(
                torch.cat([self.v.pos_embed[:, :2, :].detach(), new_pos_embed], dim=1)
            )

    def get_shape(self, fstride, tstride, input_fdim=128, input_tdim=1024):
        test_input = torch.randn(1, 1, input_fdim, input_tdim)
        test_proj = nn.Conv2d(
            1,
            self.original_embedding_dim,
            kernel_size=(16, 16),
            stride=(fstride, tstride),
        )
        test_out = test_proj(test_input)
        f_dim = test_out.shape[2]
        t_dim = test_out.shape[3]
        return f_dim, t_dim

    @autocast()
    def forward(self, x):
        """
        :param x: the input spectrogram, expected shape: (batch_size, time_frame_num, frequency_bins), e.g., (12, 1024, 128)
        :return: prediction
        """
        # expect input x = (batch_size, time_frame_num, frequency_bins), e.g., (12, 1024, 128)
        x = x.unsqueeze(1)
        x = x.transpose(2, 3)

        B = x.shape[0]
        x = self.v.patch_embed(x)
        cls_tokens = self.v.cls_token.expand(B, -1, -1)
        dist_token = self.v.dist_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, dist_token, x), dim=1)
        x = x + self.v.pos_embed
        x = self.v.pos_drop(x)

        hidden_states = []
        for blk in self.v.blocks:
            x = blk(x)
            post_x = x[
                :, 2:
            ]  # (batch_size, num_vertical_patch * num_horizon_patch, hidden_size)
            post_x = (
                post_x.reshape(B, NUM_VERTICAL_PATCH, -1, x.size(-1))
                .transpose(1, 2)
                .flatten(2)
            )
            # (batch_size, num_horizon_patch, num_vertical_patch * hidden_size)
            hidden_states.append(post_x)

        x = self.v.norm(x)
        x = (x[:, 0] + x[:, 1]) / 2

        x = self.mlp_head(x)  # (batch_size, num_class)
        return x, hidden_states



"""
Benchmark the timing a block of code

Authors
  * Leo 2022
"""

import logging
from collections import defaultdict
from contextlib import ContextDecorator
from time import time
from typing import Any

import numpy as np
import torch

logger = logging.getLogger(__name__)
_history = defaultdict(list)

__all__ = ["benchmark"]


class benchmark(ContextDecorator):
    def __init__(self, name: str, freq: int = 1) -> None:
        super().__init__()
        self.name = name
        self.freq = freq

    def __enter__(self):
        torch.cuda.synchronize()
        self.start = time()

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        torch.cuda.synchronize()
        seconds = time() - self.start

        global _history
        _history[self.name].append(seconds)
        if len(_history[self.name]) % self.freq == 0:
            logger.warning(
                f"{self.name}: {seconds} secs, avg {np.array(_history[self.name]).mean()} secs"
            )



"""
Parse the Audio SNIPS corpus

Authors:
  * Heng-Jui Chang 2022
"""

import logging
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List

from tqdm import trange

from .base import Corpus

__all__ = [
    "SNIPS",
]


class SNIPS(Corpus):
    def __init__(
        self,
        dataset_root: str,
        train_speakers: List[str],
        valid_speakers: List[str],
        test_speakers: List[str],
    ) -> None:
        self.dataset_root = Path(dataset_root)
        self.train_speakers = train_speakers
        self.valid_speakers = valid_speakers
        self.test_speakers = test_speakers

        self.data_dict = self._collect_data(
            self.dataset_root, train_speakers, valid_speakers, test_speakers
        )
        self.train = self._data_to_dict(self.data_dict, ["train"])
        self.valid = self._data_to_dict(self.data_dict, ["valid"])
        self.test = self._data_to_dict(self.data_dict, ["test"])

        self._data = OrderedDict()
        self._data.update(self.train)
        self._data.update(self.valid)
        self._data.update(self.test)

    @property
    def all_data(self):
        return self._data

    @property
    def data_split_ids(self):
        return (
            list(self.train.keys()),
            list(self.valid.keys()),
            list(self.test.keys()),
        )

    @staticmethod
    def _collect_data(
        dataset_root: str,
        train_speakers: List[str],
        valid_speakers: List[str],
        test_speakers: List[str],
    ) -> Dict[str, Dict[str, Any]]:
        # Load transcription
        transcripts_file = open(dataset_root / "all.iob.snips.txt").readlines()
        transcripts = {}
        for line in transcripts_file:
            line = line.strip().split(" ")
            index = line[0]  # {speaker}-snips-{split}-{index}
            sent = " ".join(line[1:])
            transcripts[index] = sent

        # List wave files
        data_dict = {}
        for split, speaker_list in [
            ("train", train_speakers),
            ("valid", valid_speakers),
            ("test", test_speakers),
        ]:
            wav_list = list((dataset_root / split).rglob("*.wav"))
            new_wav_list, name_list, spkr_list = [], [], []
            uf = 0
            for i in trange(len(wav_list), desc="checking files"):
                uid = wav_list[i].stem
                if uid in transcripts:
                    spkr = uid.split("-")[0]
                    if spkr in speaker_list:
                        new_wav_list.append(str(wav_list[i]))
                        name_list.append(uid)
                        spkr_list.append(spkr)
                else:
                    logging.info(wav_list[i], "Not Found")
                    uf += 1

            logging.info("%d wav file with label not found in text file!" % uf)
            wav_list = new_wav_list
            logging.info(
                f"loaded audio from {len(speaker_list)} speakers {str(speaker_list)} with {len(wav_list)} examples."
            )
            assert len(wav_list) > 0, "No data found @ {}".format(dataset_root / split)

            text_list = [transcripts[name] for name in name_list]

            wav_list, name_list, text_list, spkr_list = zip(
                *[
                    (wav, name, text, spkr)
                    for (wav, name, text, spkr) in sorted(
                        zip(wav_list, name_list, text_list, spkr_list),
                        key=lambda x: x[1],
                    )
                ]
            )

            data_dict[split] = {
                "name_list": name_list,
                "wav_list": wav_list,
                "text_list": text_list,
                "spkr_list": spkr_list,
            }

        return data_dict

    @staticmethod
    def _data_to_dict(
        data_dict: Dict[str, Dict[str, List[Any]]], splits: List[str]
    ) -> dict:
        data = dict(
            {
                name: {
                    "wav_path": data_dict[split]["wav_list"][i],
                    "transcription": " ".join(
                        data_dict[split]["text_list"][i]
                        .split("\t")[0]
                        .strip()
                        .split(" ")[1:-1]
                    ),
                    "iob": " ".join(
                        data_dict[split]["text_list"][i]
                        .split("\t")[1]
                        .strip()
                        .split(" ")[1:-1]
                    ),
                    "intent": data_dict[split]["text_list"][i]
                    .split("\t")[1]
                    .strip()
                    .split(" ")[-1],
                    "speaker": data_dict[split]["spkr_list"][i],
                    "corpus_split": split,
                }
                for split in splits
                for i, name in enumerate(data_dict[split]["name_list"])
            }
        )
        return data



"""
The setting of Superb PR

Authors
  * Heng-Jui Chang 2022
  * Leo 2022
"""

from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from omegaconf import MISSING

from s3prl.dataio.dataset import get_info
from s3prl.dataio.encoder.g2p import G2P
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler, SortedSliceSampler
from s3prl.nn.linear import FrameLevelLinear

from .superb_asr import SuperbASR

__all__ = [
    "SuperbPR",
]


class SuperbPR(SuperbASR):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                train_sets=["train-clean-100"],
                valid_sets=["dev-clean"],
                test_sets=["test-clean"],
            ),
            prepare_tokenizer_data=dict(),
            build_tokenizer=dict(
                vocab_type="phoneme",
            ),
            build_dataset=dict(),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=16,
                    max_length=300000,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_size=256,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                log_metrics=["per"],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-2,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(
                extra_conf=dict(
                    build_downstream_conf="${build_downstream}"
                ),  # This is redundant for ASR. Just to show how to clone other fields
            ),
            save_task=dict(),
            train=dict(
                total_steps=100000,
                log_step=100,
                eval_step=1000,
                save_step=100,
                gradient_clipping=1.0,
                gradient_accumulate=2,
                valid_metric="per",
                valid_higher_better=False,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        train_csv, valid_csv, test_csvs = super().prepare_data(
            prepare_data, target_dir, cache_dir, get_path_only
        )
        if get_path_only:
            return train_csv, valid_csv, test_csvs

        g2p = G2P()

        def phonemize(csv_path):
            df = pd.read_csv(csv_path)
            text = df["transcription"].tolist()
            phonemized_text = [g2p.encode(t.strip()) for t in text]
            df["transcription"] = phonemized_text
            df.to_csv(csv_path, index=False)

        for csv_path in [train_csv, valid_csv, *test_csvs]:
            phonemize(csv_path)

        return train_csv, valid_csv, test_csvs

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        """
        Return the batch sampler for torch DataLoader.

        Args:
            build_batch_sampler (dict): same in :obj:`default_config`

                ====================  ====================
                key                   description
                ====================  ====================
                train                 (dict) - arguments for :obj:`SortedSliceSampler`
                valid                 (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                test                  (dict) - arguments for :obj:`FixedBatchSizeBatchSampler`
                ====================  ====================

            target_dir (str): Current experiment directory
            cache_dir (str): If the preprocessing takes too long time, save
                the temporary files into this directory. This directory is expected to be shared
                across different training sessions (different hypers and :code:`target_dir`)
            mode (str): train/valid/test
            data_csv (str): the :code:`mode` specific csv from :obj:`prepare_data`
            dataset: the dataset from :obj:`build_dataset`

        Returns:
            batch sampler for torch DataLoader
        """

        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_batch_sampler)

        if mode == "train":
            wav_lens = get_info(dataset, ["x_len"], Path(target_dir) / "train_stats")
            sampler = SortedSliceSampler(wav_lens, **(conf.train or {}))
        elif mode == "valid":
            sampler = FixedBatchSizeBatchSampler(dataset, **(conf.valid or {}))
        elif mode == "test":
            sampler = FixedBatchSizeBatchSampler(dataset, **(conf.test or {}))

        return sampler

    def build_downstream(
        self,
        build_downstream: dict,
        downstream_input_size: int,
        downstream_output_size: int,
        downstream_input_stride: int,
    ):
        """
        Return the task-specific downstream model.
        By default build the :obj:`FrameLevelLinear`

        Args:
            build_downstream (dict): same in :obj:`default_config`,
                supports arguments in :obj:`FrameLevelLinear`
            downstream_input_size (int): the required input size of the model
            downstream_output_size (int): the required output size of the model
            downstream_input_stride (int): the input feature's stride (from 16 KHz)

        Returns:
            :obj:`s3prl.nn.interface.AbsFrameModel`
        """
        return FrameLevelLinear(
            downstream_input_size, downstream_output_size, **build_downstream
        )



import logging

import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence

from .lighthubert import LightHuBERT, LightHuBERTConfig

logger = logging.getLogger(__name__)


class UpstreamExpert(torch.nn.Module):
    def __init__(self, ckpt, **kwds):
        super().__init__()

        checkpoint = torch.load(ckpt)
        assert checkpoint["cfg"]["model"]["_name"] in [
            "hubert_pruner",
            "student_hubert",
        ]
        self.cfg = LightHuBERTConfig(checkpoint["cfg"]["model"])

        if checkpoint["cfg"]["model"]["_name"] == "hubert_pruner":
            if (
                checkpoint["cfg"]["model"]["pruner_supernet"]
                .lower()
                .endswith("small.yaml")
            ):
                self.cfg.supernet_type = "small"
            elif (
                checkpoint["cfg"]["model"]["pruner_supernet"]
                .lower()
                .endswith("base.yaml")
            ):
                self.cfg.supernet_type = "base"

        self.model = LightHuBERT(self.cfg)
        self.model.load_state_dict(checkpoint["model"], strict=False)

        if checkpoint["cfg"]["model"]["_name"] == "student_hubert":
            subnet = self.model.supernet.max_subnet
        else:
            subnet = self.model.supernet.subnet
        self.model.set_sample_config(subnet)

        self.model.encoder.layerdrop = 0.0

        params = self.model.calc_sampled_param_num()
        logger.info(f"LightHubert subnet (Params {params / 1e6:.0f}M) | {subnet}")

    def get_downsample_rates(self, key: str) -> int:
        return 320

    def forward(self, wavs):
        wavs = [F.layer_norm(wav, wav.shape) for wav in wavs]

        device = wavs[0].device
        wav_lengths = torch.LongTensor([len(wav) for wav in wavs]).to(device)
        wav_padding_mask = ~torch.lt(
            torch.arange(max(wav_lengths)).unsqueeze(0).to(device),
            wav_lengths.unsqueeze(1),
        )

        hs = self.model.extract_features(
            pad_sequence(wavs, batch_first=True),
            padding_mask=wav_padding_mask,
            ret_hs=True,
        )[0]

        return {
            "hidden_states": hs,
        }



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/vggish/hubconf.py ]
#   Synopsis     [ the VGGish torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import torch

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def _load_state_dict_from_url(url):
    path = _urls_to_filepaths(url)
    return torch.load(path, map_location="cpu")


def _vggish_from_torch_hub(urls, *args, **kwargs):
    """
    The model from `torch.hub.load`
        urls (dict): LINKS
    """
    kwargs["ckpt"] = {
        "vggish": _load_state_dict_from_url(urls["vggish"]),
        "pca": _load_state_dict_from_url(urls["pca"]),
    }
    return _UpstreamExpert(*args, **kwargs)


def vggish(*args, **kwargs):
    """
    The default model
    """
    urls = {
        "vggish": "https://github.com/harritaylor/torchvggish/releases/download/v0.1/vggish-10086976.pth",
        "pca": "https://github.com/harritaylor/torchvggish/releases/download/v0.1/vggish_pca_params-970ea276.pth",
    }
    return _vggish_from_torch_hub(urls, *args, **kwargs)



from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def decoar_custom(ckpt: str, refresh=False, *args, **kwargs):
    if ckpt.startswith("http"):
        ckpt = _urls_to_filepaths(ckpt, refresh=refresh)

    return _UpstreamExpert(ckpt, *args, **kwargs)


def decoar_local(*args, **kwargs):
    return decoar_custom(*args, **kwargs)


def decoar_url(*args, **kwargs):
    return decoar_custom(*args, **kwargs)


def decoar(refresh=False, *args, **kwargs):
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/checkpoint_decoar.pt"
    return decoar_url(*args, refresh=refresh, **kwargs)



import numpy as np
import torch
from torch.nn.utils.rnn import pad_sequence

__all__ = [
    "default_collate_fn",
]


def default_collate_fn(samples, padding_value: int = 0):
    """
    Each item in **DynamicItemDataset** is a dict
    This function pad (or transform into numpy list) a batch of dict

    Args:
        samples (List[dict]): Suppose each Container is in

            .. code-block:: yaml

                wav: a single waveform
                label: a single string

    Return:
        dict

        .. code-block:: yaml

            wav: padded waveforms
            label: np.array([a list of string labels])
    """
    assert isinstance(samples[0], dict)
    keys = samples[0].keys()
    padded_samples = dict()
    for key in keys:
        values = [sample[key] for sample in samples]
        if isinstance(values[0], int):
            values = torch.LongTensor(values)
        elif isinstance(values[0], float):
            values = torch.FloatTensor(values)
        elif isinstance(values[0], np.ndarray):
            values = [torch.from_numpy(value).float() for value in values]
            values = pad_sequence(values, batch_first=True, padding_value=padding_value)
        elif isinstance(values[0], torch.Tensor):
            values = pad_sequence(values, batch_first=True, padding_value=padding_value)
        else:
            values = np.array(values, dtype="object")
        padded_samples[key] = values
    return padded_samples



"""
Parse the LibriLight corpus

Authors:
  * Heng-Jui Chang 2022
"""

import logging
import os
import re
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List

from joblib import Parallel, delayed

from s3prl.util.download import _urls_to_filepaths

from .base import Corpus

LIBRILIGHT_SPLITS = [
    "10h",
    "1h",
    "10m-fold0",
    "10m-fold1",
    "10m-fold2",
    "10m-fold3",
    "10m-fold4",
    "10m-fold5",
]
LIBRISPEECH_SPKR_INFO = (
    "https://huggingface.co/datasets/s3prl/librispeech_metadata/raw/main/SPEAKERS.TXT"
)


__all__ = [
    "LibriLight",
]


def read_text(file: Path) -> str:
    src_file = "-".join(str(file).split("-")[:-1]) + ".trans.txt"
    idx = file.stem.replace(".flac", "")

    with open(src_file, "r") as fp:
        for line in fp:
            if idx == line.split(" ")[0]:
                return line[:-1].split(" ", 1)[1]

    logging.warning(f"Transcription of {file} not found!")


def check_no_repeat(splits: List[str]) -> bool:
    count = defaultdict(int)
    for split in splits:
        count[split] += 1

    repeated = ""
    for key, val in count.items():
        if val > 1:
            repeated += f" {key} ({val} times)"

    if len(repeated) != 0:
        logging.warning(
            f"Found repeated splits in corpus: {repeated}, which might cause unexpected behaviors."
        )
        return False

    return True


def _parse_spk_to_gender(speaker_file: Path) -> dict:
    speaker_file = Path(speaker_file)
    with speaker_file.open() as file:
        lines = [line.strip() for line in file.readlines()]
    for line_id in range(len(lines)):
        line = lines[line_id]
        if "SEX" in line and "SUBSET" in line and "MINUTES" in line and "NAME" in line:
            break

    line_id += 1  # first line with speaker info
    spk2gender = {}
    for line_id in range(line_id, len(lines)):
        line = lines[line_id]
        line = re.sub("\t+", " ", line)
        line = re.sub(" +", " ", line)
        parts = line.split("|", maxsplit=4)
        ID, SEX, SUBSET, MINUTES, NAME = parts
        spk2gender[int(ID)] = SEX.strip()
    return spk2gender


class LibriLight(Corpus):
    def __init__(
        self,
        dataset_root: str,
        n_jobs: int = 4,
        train_split: str = "10m-fold0",
    ) -> None:
        self.dataset_root = Path(dataset_root).resolve()
        self.train_split = train_split

        if train_split == "10h":
            roots = [self.dataset_root / "1h", self.dataset_root / "9h"]
        elif train_split == "1h":
            roots = [self.dataset_root / "1h"]
        elif train_split.startswith("10m"):
            fold_id = int(train_split.split("-")[-1].split("fold")[-1])
            roots = [self.dataset_root / "1h" / str(fold_id)]
        else:
            raise ValueError(f"Unsupported split: {train_split}")

        self._data = self._collect_data(roots, n_jobs)

    @classmethod
    def download_dataset(cls, dataset_root: str):
        Path(dataset_root).mkdir(parents=True, exist_ok=True)
        subprocess.check_call(
            [
                "wget",
                "https://dl.fbaipublicfiles.com/librilight/data/librispeech_finetuning.tgz",
                "-O",
                str(Path(dataset_root) / "librispeech_finetuning.tgz"),
            ]
        )
        subprocess.check_call(
            ["tar", "zxvf", "librispeech_finetuning.tgz", "-C", str(Path(dataset_root))]
        )

    @property
    def all_data(self):
        return self._data

    @staticmethod
    def _collect_data(
        roots: List[Path], n_jobs: int = 4
    ) -> Dict[str, Dict[str, List[Any]]]:
        spkr_file = _urls_to_filepaths(LIBRISPEECH_SPKR_INFO)
        spkr2gender = _parse_spk_to_gender(Path(spkr_file).resolve())
        data_dict = {}
        for split_dir in roots:
            if not os.path.exists(split_dir):
                logging.info(
                    f"Split {split_dir} is not downloaded. Skip data collection."
                )
                continue

            wav_list = list(Path(split_dir).rglob("*.flac"))
            name_list = [file.stem.replace(".flac", "") for file in wav_list]
            text_list = Parallel(n_jobs=n_jobs)(
                delayed(read_text)(file) for file in wav_list
            )
            spkr_list = [int(name.split("-")[0]) for name in name_list]

            for wav_id in range(len(wav_list)):
                wav = Path(wav_list[wav_id])
                data_dict[wav.stem] = {
                    "wav_path": str(wav.resolve()),
                    "transcription": text_list[wav_id],
                    "speaker": spkr_list[wav_id],
                    "gender": spkr2gender[spkr_list[wav_id]],
                }
        return data_dict



"""
Common model and loss in pure torch.nn.Module with torch dependency only
"""

from .beam_decoder import BeamDecoder
from .common import FrameLevel, UtteranceLevel
from .linear import FrameLevelLinear, MeanPoolingLinear
from .pooling import MeanPooling, TemporalAveragePooling, TemporalStatisticsPooling
from .rnn import RNNEncoder, SuperbDiarizationModel
from .speaker_loss import amsoftmax, softmax
from .speaker_model import SuperbXvector, XVectorBackbone
from .upstream import Featurizer, S3PRLUpstream

__all__ = [
    "S3PRLUpstream",
    "Featurizer",
    "FrameLevel",
    "UtteranceLevel",
    "FrameLevelLinear",
    "MeanPoolingLinear",
    "MeanPooling",
    "TemporalAveragePooling",
    "TemporalStatisticsPooling",
    "RNNEncoder",
    "SuperbDiarizationModel",
    "amsoftmax",
    "softmax",
    "SuperbXvector",
    "XVectorBackbone",
    "BeamDecoder",
]



"""
The beam search decoder of flashlight

Authors:
  * Heng-Jui Chang 2022
"""

import itertools as it
import logging
import math
from typing import Iterable, List

import torch

from s3prl.util.download import _urls_to_filepaths

TOKEN_URL = "https://huggingface.co/datasets/s3prl/flashlight/raw/main/lexicon/librispeech_char_tokens.txt"

LEXICON_URL_1 = "https://dl.fbaipublicfiles.com/fairseq/wav2vec/librispeech_lexicon.lst"
LEXICON_URL_2 = "https://huggingface.co/datasets/s3prl/flashlight/raw/main/lexicon/librispeech_lexicon.lst"

LM_URL_1 = "https://www.openslr.org/resources/11/4-gram.arpa.gz"
LM_URL_2 = (
    "https://huggingface.co/datasets/s3prl/flashlight/resolve/main/lm/4-gram.arpa.gz"
)

logger = logging.getLogger(__name__)

__all__ = ["BeamDecoder"]


class BeamDecoder(object):
    """Beam decoder powered by flashlight.

    Args:
        token (str, optional): Path to dictionary file. Defaults to "".
        lexicon (str, optional): Path to lexicon file. Defaults to "".
        lm (str, optional): Path to KenLM file. Defaults to "".
        nbest (int, optional): Returns nbest hypotheses. Defaults to 1.
        beam (int, optional): Beam size. Defaults to 5.
        beam_size_token (int, optional): Token beam size. Defaults to -1.
        beam_threshold (float, optional): Beam search log prob threshold. Defaults to 25.0.
        lm_weight (float, optional): language model weight. Defaults to 2.0.
        word_score (float, optional): score for words appearance in the transcription. Defaults to -1.0.
        unk_score (float, optional): score for unknown word appearance in the transcription. Defaults to -math.inf.
        sil_score (float, optional): score for silence appearance in the transcription. Defaults to 0.0.
    """

    def __init__(
        self,
        token: str = "",
        lexicon: str = "",
        lm: str = "",
        nbest: int = 1,
        beam: int = 5,
        beam_size_token: int = -1,
        beam_threshold: float = 25.0,
        lm_weight: float = 2.0,
        word_score: float = -1.0,
        unk_score: float = -math.inf,
        sil_score: float = 0.0,
    ):
        try:
            from flashlight.lib.text.decoder import (
                CriterionType,
                KenLM,
                LexiconDecoder,
                LexiconDecoderOptions,
                SmearingMode,
                Trie,
            )
            from flashlight.lib.text.dictionary import (
                Dictionary,
                create_word_dict,
                load_words,
            )

        except ImportError:
            logger.error(
                f"Please install Flashlight Text from https://github.com/flashlight/text to enable {__class__.__name__}"
            )
            raise

        if token == "":
            token = _urls_to_filepaths(TOKEN_URL)
        if lexicon == "":
            # Try LEXICON_URL_2 if LEXICON_URL_1 did not work.
            lexicon = _urls_to_filepaths(LEXICON_URL_1)
        if lm == "":
            # Try LM_URL_2 if LM_URL_1 did not work.
            lm = _urls_to_filepaths(LM_URL_1)

        self.nbest = nbest

        self.token_dict = Dictionary(token)
        self.lexicon = load_words(lexicon)
        self.word_dict = create_word_dict(self.lexicon)

        self.lm = KenLM(lm, self.word_dict)

        self.sil_idx = self.token_dict.get_index("|")
        self.unk_idx = self.word_dict.get_index("<unk>")

        self.trie = Trie(self.token_dict.index_size(), self.sil_idx)
        start_state = self.lm.start(False)

        for word, spellings in self.lexicon.items():
            usr_idx = self.word_dict.get_index(word)
            _, score = self.lm.score(start_state, usr_idx)
            for spelling in spellings:
                spelling_idxs = [self.token_dict.get_index(tok) for tok in spelling]
                self.trie.insert(spelling_idxs, usr_idx, score)
        self.trie.smear(SmearingMode.MAX)

        if beam_size_token == -1:
            beam_size_token = self.token_dict.index_size()

        self.options = LexiconDecoderOptions(
            beam_size=beam,
            beam_size_token=beam_size_token,
            beam_threshold=beam_threshold,
            lm_weight=lm_weight,
            word_score=word_score,
            unk_score=unk_score,
            sil_score=sil_score,
            log_add=False,
            criterion_type=CriterionType.CTC,
        )

        self.blank_idx = self.token_dict.get_index("#")
        self.decoder = LexiconDecoder(
            self.options,
            self.trie,
            self.lm,
            self.sil_idx,
            self.blank_idx,
            self.unk_idx,
            [],
            False,
        )

    def get_tokens(self, idxs: Iterable) -> torch.LongTensor:
        """Normalize tokens by handling CTC blank, ASG replabels, etc.

        Args:
            idxs (Iterable): Token ID list output by self.decoder

        Returns:
            torch.LongTensor: Token ID list after normalization.
        """

        idxs = (g[0] for g in it.groupby(idxs))
        idxs = filter(lambda x: x != self.blank_idx, idxs)
        return torch.LongTensor(list(idxs))

    def get_timesteps(self, token_idxs: List[int]) -> List[int]:
        """Returns frame numbers corresponding to every non-blank token.

        Args:
            token_idxs (List[int]): IDs of decoded tokens.

        Returns:
            List[int]: Frame numbers corresponding to every non-blank token.
        """

        timesteps = []
        for i, token_idx in enumerate(token_idxs):
            if token_idx == self.blank_idx:
                continue
            if i == 0 or token_idx != token_idxs[i - 1]:
                timesteps.append(i)
        return timesteps

    def decode(self, emissions: torch.Tensor) -> List[List[dict]]:
        """Decode sequence.

        Args:
            emissions (torch.Tensor): Emission probabilities (in log scale).

        Returns:
            List[List[dict]]: Decoded hypotheses.
        """

        emissions = emissions.float().contiguous().cpu()
        B, T, N = emissions.size()

        hyps = []
        for b in range(B):
            emissions_ptr = emissions.data_ptr() + 4 * b * emissions.stride(0)
            results = self.decoder.decode(emissions_ptr, T, N)
            nbest_results = results[: self.nbest]
            hyps.append(
                [
                    dict(
                        tokens=self.get_tokens(result.tokens),
                        score=result.score,
                        timesteps=self.get_timesteps(result.tokens),
                        words=[
                            self.word_dict.get_entry(x) for x in result.words if x >= 0
                        ],
                    )
                    for result in nbest_results
                ]
            )
        return hyps



"""
The most commonly used batch sampler in S3PRL legacy codebase,
which sorts the lengths of all the data points and group the instances
with the similar lengths together.

Authors:
  Leo 2022
"""

from typing import List

import torch

__all__ = [
    "SortedSliceSampler",
    "SortedBucketingSampler",
]


class SortedSliceSampler:
    """
    This sampler should only be used for training hence is always in random shuffle mode

    Args:
        lengths (List[int])
        batch_size (int): the default batch size
        max_length (int): if a batch contains at least on utt longer than max_length, half the batch
        get_length_func (callable): get the length of each item in the dataset, if None, a default function will be used
        in_batch_shuffle (bool): if False, batches are sorted by length from long to short
    """

    def __init__(
        self,
        lengths: List[int],
        batch_size: int,
        max_length: int = 300000,
        seed: int = 12345678,
        in_batch_shuffle: bool = False,
    ) -> None:
        self.lengths = lengths
        self.epoch = 0
        self.seed = seed
        self.batch_size = batch_size
        self.max_length = max_length
        self.in_batch_shuffle = in_batch_shuffle

        sorted_ids = [(idx, length) for idx, length in enumerate(lengths)]
        sorted_ids = sorted(sorted_ids, key=lambda x: x[1], reverse=True)
        self.sorted_ids = [data_id for data_id, length in sorted_ids]

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def __iter__(self):
        generator = torch.Generator()
        generator.manual_seed(self.epoch + self.seed)

        indices = torch.randperm(len(self.lengths), generator=generator).tolist()

        for indice in indices:
            length = self.lengths[indice]
            if length > self.max_length:
                batch_size = self.batch_size // 2
            else:
                batch_size = self.batch_size
            start_position = self.sorted_ids.index(indice)
            batch = self.sorted_ids[start_position : start_position + batch_size]

            if self.in_batch_shuffle:
                inbatch_indices = torch.randperm(
                    len(batch), generator=generator
                ).tolist()
                batch = [batch[idx] for idx in inbatch_indices]

            yield batch

    def __len__(self):
        return len(list(iter(self)))


class SortedBucketingSampler:
    """
    Args:
        lengths (List[int])
        batch_size (int): the default batch size
        max_length (int): if a batch contains at least on utt longer than max_length, half the batch
        get_length_func (callable): get the length of each item in the dataset, if None, a default function will be used
        shuffle (bool): Whether to shuffle the batches
        in_batch_shuffle (bool): if False, batches are sorted by length from long to short
    """

    def __init__(
        self,
        lengths: List[int],
        batch_size: int,
        max_length: int = 300000,
        shuffle: bool = False,
        in_batch_shuffle: bool = False,
        seed: int = 12345678,
    ) -> None:
        self.epoch = 0
        self.seed = seed
        self.batch_size = batch_size
        self.max_length = max_length
        self.shuffle = shuffle
        self.in_batch_shuffle = in_batch_shuffle
        self.lengths = lengths

        sorted_ids = [(idx, length) for idx, length in enumerate(self.lengths)]
        sorted_ids = sorted(sorted_ids, key=lambda x: x[1], reverse=True)
        self.sorted_ids = [data_id for data_id, length in sorted_ids]

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def __iter__(self):
        generator = torch.Generator()
        generator.manual_seed(self.epoch + self.seed)

        batches = []
        position = 0
        while position < len(self.sorted_ids):
            indice = self.sorted_ids[position]
            length = self.lengths[indice]
            if length > self.max_length:
                batch_size = self.batch_size // 2
            else:
                batch_size = self.batch_size
            batch = self.sorted_ids[
                position : min(position + batch_size, len(self.sorted_ids))
            ]
            position += batch_size
            if self.in_batch_shuffle:
                shuffled_batch_indices = torch.randperm(len(batch), generator=generator)
                batch = [batch[idx] for idx in shuffled_batch_indices]
            batches.append(batch)

        if self.shuffle:
            shuffled_indices = torch.randperm(len(batches), generator=generator)
            batches = [batches[idx] for idx in shuffled_indices]

        return iter(batches)

    def __len__(self):
        return len(list(iter(self)))



from pathlib import Path

try:
    # HACK: SummaryWriter must be imported at the begining or else it will lead to core dumped
    # This is a known issue: https://github.com/pytorch/pytorch/issues/30651
    from torch.utils.tensorboard.writer import SummaryWriter
except:
    pass


with (Path(__file__).parent.resolve() / "version.txt").open() as file:
    __version__ = file.read().strip()



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class SLinear(nn.Linear):
    r"""Linear Layer: variable input and output size.
    `in_splits` and `out_splits` enable equally seperate splits
        at in and out dimensions, e.g.,
        A tensor of 128 size has 4 splits with the split [0,31],
        the split [32,63], the split [64,95], and the split
        [96,127].
        If given 2.5 splits and total 40 sizes, we can conclude
        that the split [0-15], the split [32-47], and the split
        [64-71].

    __base__: torch.nn.Linear
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        in_splits: int = 1,
        out_splits: int = 1,
    ) -> None:
        super(SLinear, self).__init__(in_features, out_features, bias)
        self.in_splits = in_splits
        self.out_splits = out_splits
        assert self.in_features % self.in_splits == 0
        assert self.out_features % self.out_splits == 0
        self.in_size_split = self.in_features // self.in_splits
        self.out_size_split = self.out_features // self.out_splits
        self.staticize()

    def staticize(self):
        self.sample_in_dim = None
        self.sample_out_dim = None
        self.sample_in_splits = None
        self.sample_out_splits = None
        self.samples = {
            "weight": self.weight,
            "bias": self.bias,
        }

    def set_sample_config(
        self, sample_in_dim, sample_out_dim, sample_in_splits=1, sample_out_splits=1
    ):
        assert sample_in_dim is not None and sample_out_dim is not None
        if sample_in_splits is not None:
            assert sample_in_splits <= self.in_splits, f"{sample_in_splits}"
            assert (
                sample_in_dim // sample_in_splits > 0
            ), f"{sample_in_dim} // {sample_in_splits} <= 0"
        if sample_out_splits is not None:
            assert sample_out_splits <= self.out_splits, f"{sample_out_splits}"
            assert (
                sample_out_dim // sample_out_splits > 0
            ), f"{sample_out_dim} // {sample_out_splits} <= 0"
        self.sample_in_dim = sample_in_dim
        self.sample_out_dim = sample_out_dim
        self.sample_in_splits = sample_in_splits
        self.sample_out_splits = sample_out_splits
        self._sample_parameters()

    def _sample_parameters(self):
        if (
            self.in_splits == 1
            or self.sample_in_dim == self.in_features
            or self.sample_in_splits is None
        ):
            weight = self.weight[:, : self.sample_in_dim].contiguous()
        else:
            assert self.sample_in_splits is not None, f"{self.sample_in_splits}"
            weight_splits = torch.split(self.weight, self.in_size_split, dim=1)
            size_split = int(self.sample_in_dim / self.sample_in_splits)
            weight = torch.cat(
                [
                    split_i[:, : min(size_split, self.sample_in_dim - i * size_split)]
                    for i, split_i in enumerate(
                        weight_splits[: int(np.ceil(self.sample_in_splits))]
                    )
                ],
                dim=1,
            )

        if (
            self.out_splits == 1
            or self.sample_out_dim == self.out_features
            or self.sample_out_splits is None
        ):
            weight = weight[: self.sample_out_dim].contiguous()
            if self.bias is not None:
                bias = self.bias[: self.sample_out_dim].contiguous()
        else:
            assert self.sample_out_splits is not None, f"{self.sample_out_splits}"
            weight_splits = torch.split(weight, self.out_size_split, dim=0)
            size_split = int(self.sample_out_dim / self.sample_out_splits)
            weight = torch.cat(
                [
                    split_i[: min(size_split, self.sample_out_dim - i * size_split)]
                    for i, split_i in enumerate(
                        weight_splits[: int(np.ceil(self.sample_out_splits))]
                    )
                ],
                dim=0,
            )
            if self.bias is not None:
                bias_splits = torch.split(self.bias, self.out_size_split, dim=0)
                bias = torch.cat(
                    [
                        split_i[: min(size_split, self.sample_out_dim - i * size_split)]
                        for i, split_i in enumerate(
                            bias_splits[: int(np.ceil(self.sample_out_splits))]
                        )
                    ],
                    dim=0,
                )

        self.samples["weight"] = weight
        if self.bias is not None:
            self.samples["bias"] = bias

        # self.samples["weight"] = self.weight[:self.sample_out_dim, :self.sample_in_dim]
        # if self.bias is not None:
        #     self.samples["bias"] = self.bias[:self.sample_out_dim]
        return self.samples

    def calc_sampled_param_num(self):
        weight_numel = self.samples["weight"].numel()

        if self.samples["bias"] is not None:
            bias_numel = self.samples["bias"].numel()
        else:
            bias_numel = 0

        return weight_numel + bias_numel

    def get_complexity(self, sequence_length):
        total_flops = 0
        total_flops += sequence_length * np.prod(self.samples["weight"].size())
        return total_flops

    @property
    def weights(self):
        return self.samples["weight"]

    @property
    def biases(self):
        return self.samples["bias"]

    def forward(self, input: Tensor) -> Tensor:
        self._sample_parameters()
        return F.linear(input, self.weights, self.biases)

    def extra_repr(self) -> str:
        return "in_features={}, out_features={}, bias={}".format(
            self.in_features, self.out_features, self.bias is not None
        )

    def clone_model(
        self, in_dim: int, out_dim: int, in_splits: int = 1, out_splits: int = 1
    ):
        self.set_sample_config(in_dim, out_dim, in_splits, out_splits)

        isbias = self.bias is not None
        m = nn.Linear(in_dim, out_dim, isbias)
        m = m.to(self.weight.device)
        m = m.to(self.weight.dtype)
        m.weight.data.copy_(self.weights)
        if isbias:
            m.bias.data.copy_(self.biases)
        return m.eval()

    @classmethod
    def build_from(cls, m: nn.Linear, in_splits: int = 1, out_splits: int = 1):
        in_features = m.in_features
        out_features = m.out_features
        isbias = m.bias is not None
        _m = cls(in_features, out_features, isbias, in_splits, out_splits)
        _m = _m.to(m.weight.device)
        _m = _m.to(m.weight.dtype)
        _m.weight.data.copy_(m.weight)
        if isbias:
            _m.bias.data.copy_(m.bias)
        return _m


if __name__ == "__main__":
    print(SLinear.__base__)
    m = SLinear(2, 3)
    z = m.clone_model(2, 2)
    x = SLinear.build_from(z)
    print(m, z, x)
    inp = torch.rand((1, 2))
    m.set_sample_config(2, 2)
    print(torch.allclose(m(inp), z(inp)))
    print(torch.allclose(m(inp), x(inp)))

    m = SLinear(2, 4, in_splits=2, out_splits=4)
    z = m.clone_model(2, 2, 2, 2)
    x = SLinear.build_from(z, 2, 2)
    inp = torch.rand((1, 2))
    m.set_sample_config(2, 2, 2, 2)
    print(torch.allclose(m(inp), z(inp)))
    print(torch.allclose(m(inp), x(inp)))



"""BYOL for Audio

Kudos to Phil Wang, this implementation is based on https://github.com/lucidrains/byol-pytorch/

This code is customized to enable:
- Decoupling augmentations.
- Feeding two augmented input batches independently.
"""

import copy
import random
from functools import wraps

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

# helper functions


def default(val, def_val):
    return def_val if val is None else val


def flatten(t):
    return t.reshape(t.shape[0], -1)


def singleton(cache_key):
    def inner_fn(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            instance = getattr(self, cache_key)
            if instance is not None:
                return instance

            instance = fn(self, *args, **kwargs)
            setattr(self, cache_key, instance)
            return instance

        return wrapper

    return inner_fn


def get_module_device(module):
    return next(module.parameters()).device


def set_requires_grad(model, val):
    for p in model.parameters():
        p.requires_grad = val


# loss fn


def loss_fn(x, y):
    x = F.normalize(x, dim=-1, p=2)
    y = F.normalize(y, dim=-1, p=2)
    return 2 - 2 * (x * y).sum(dim=-1)


# exponential moving average


class EMA:
    def __init__(self, beta):
        super().__init__()
        self.beta = beta

    def update_average(self, old, new):
        if old is None:
            return new
        return old * self.beta + (1 - self.beta) * new


def update_moving_average(ema_updater, ma_model, current_model):
    for current_params, ma_params in zip(
        current_model.parameters(), ma_model.parameters()
    ):
        old_weight, up_weight = ma_params.data, current_params.data
        ma_params.data = ema_updater.update_average(old_weight, up_weight)


# MLP class for projector and predictor


class MLP(nn.Module):
    def __init__(self, dim, projection_size, hidden_size=4096, use_bn=True):
        super().__init__()

        self.lin1 = nn.Linear(dim, hidden_size)
        self.lin2 = nn.Linear(hidden_size, projection_size)

        self.use_bn = use_bn

        self.bn = nn.BatchNorm1d(hidden_size)
        self.relu = nn.ReLU(inplace=True)

        """self.net = nn.Sequential(
            nn.Linear(dim, hidden_size),
            nn.BatchNorm1d(hidden_size),
            nn.ReLU(inplace=True),
            nn.Linear(hidden_size, projection_size)
        )"""

    def forward(self, x):
        x = self.lin1(x)
        if self.use_bn:
            x = self.bn(x)
        x = self.relu(x)
        x = self.lin2(x)
        return x


# a wrapper class for the base neural network
# will manage the interception of the hidden layer output
# and pipe it into the projecter and predictor nets


class NetWrapper(nn.Module):
    def __init__(self, net, projection_size, projection_hidden_size, layer=-2):
        super().__init__()
        self.net = net
        self.layer = layer

        self.projector = None
        self.projection_size = projection_size
        self.projection_hidden_size = projection_hidden_size

        self.hidden = {}
        self.hook_registered = False

    def _find_layer(self):
        if type(self.layer) == str:
            modules = dict([*self.net.named_modules()])
            return modules.get(self.layer, None)
        elif type(self.layer) == int:
            children = [*self.net.children()]
            return children[self.layer]
        return None

    def _hook(self, _, input, output):
        device = input[0].device
        self.hidden[device] = flatten(output)

    def _register_hook(self):
        layer = self._find_layer()
        assert layer is not None, f"hidden layer ({self.layer}) not found"
        handle = layer.register_forward_hook(self._hook)
        self.hook_registered = True

    @singleton("projector")
    def _get_projector(self, hidden):
        # print(hidden.shape)
        # assert False
        _, dim = hidden.shape
        projector = MLP(dim, self.projection_size, self.projection_hidden_size)
        return projector.to(hidden)

    def get_representation(self, x):
        if self.layer == -1:
            return self.net(x)

        if not self.hook_registered:
            self._register_hook()

        self.hidden.clear()
        _ = self.net(x)
        hidden = self.hidden[x.device]
        self.hidden.clear()

        assert hidden is not None, f"hidden layer {self.layer} never emitted an output"
        return hidden

    def forward(self, x, return_projection=True):
        representation = self.get_representation(x)

        if not return_projection:
            return representation

        projector = self._get_projector(representation)
        projection = projector(representation)
        return projection, representation


# main class


class BYOL(nn.Module):
    """BYOL training module that is:
    - Decoupled augmentations.
    - Accepts two augmented inputs independently.
    """

    def __init__(
        self,
        net,
        image_size,
        hidden_layer=-1,
        projection_size=256,
        projection_hidden_size=4096,
        moving_average_decay=0.99,
        use_momentum=True,
        channels=1,
    ):
        super().__init__()
        self.net = net

        self.online_encoder = NetWrapper(
            net, projection_size, projection_hidden_size, layer=hidden_layer
        )

        self.use_momentum = use_momentum
        self.target_encoder = None
        self.target_ema_updater = EMA(moving_average_decay)

        self.online_predictor = MLP(
            projection_size, projection_size, projection_hidden_size
        )

        # get device of network and make wrapper same device
        device = get_module_device(net)
        self.to(device)

        # send a mock image tensor to instantiate singleton parameters
        with torch.no_grad():
            self.forward(
                torch.randn(2, channels, image_size[0], image_size[1]),
                torch.randn(2, channels, image_size[0], image_size[1]),
            )

    @singleton("target_encoder")
    def _get_target_encoder(self):
        target_encoder = copy.deepcopy(self.online_encoder)
        set_requires_grad(target_encoder, False)
        return target_encoder

    def reset_moving_average(self):
        del self.target_encoder
        self.target_encoder = None

    def update_moving_average(self):
        assert (
            self.use_momentum
        ), "you do not need to update the moving average, since you have turned off momentum for the target encoder"
        assert (
            self.target_encoder is not None
        ), "target encoder has not been created yet"
        update_moving_average(
            self.target_ema_updater, self.target_encoder, self.online_encoder
        )

    def forward(
        self, image_one, image_two, return_embedding=False, return_projection=True
    ):
        # WTF???
        # if return_embedding:
        #     return self.online_encoder(x, return_projection=return_projection)

        online_proj_one, _ = self.online_encoder(image_one)
        online_proj_two, _ = self.online_encoder(image_two)

        # My solution
        # if return_embedding:
        #     return torch.cat([online_proj_one, online_proj_two])

        online_pred_one = self.online_predictor(online_proj_one)
        online_pred_two = self.online_predictor(online_proj_two)

        with torch.no_grad():
            target_encoder = (
                self._get_target_encoder() if self.use_momentum else self.online_encoder
            )
            target_proj_one, _ = target_encoder(image_one)
            target_proj_two, _ = target_encoder(image_two)
            target_proj_one.detach_()
            target_proj_two.detach_()

        loss_one = loss_fn(online_pred_one, target_proj_two.detach())
        loss_two = loss_fn(online_pred_two, target_proj_one.detach())

        loss = loss_one + loss_two
        return loss.mean()



import sys
import traceback
from s3prl import problem


def main():
    try:
        cls = getattr(problem, sys.argv[1])
    except:
        available_problems = [
            name
            for name in dir(problem)
            if not name.startswith("_") and isinstance(getattr(problem, name), type)
        ]
        print(traceback.format_exc())
        print(
            "Usage:\n"
            "1. s3prl-main [PROBLEM] -h\n"
            "2. python3 -m s3prl.main [PROBLEM] -h\n"
            "3. python3 s3prl/main.py [PROBLEM] -h\n"
            "\nPROBLEM should be an available class name in the s3prl.problem package.\n"
            f"Available options: {', '.join(available_problems)}"
        )
        exit(0)

    cls().main(sys.argv[2:])


if __name__ == "__main__":
    main()




"""
Load tokenizer to encode & decode

Modified from tensorflow_datasets.features.text.*
Reference: https://www.tensorflow.org/datasets/api_docs/python/tfds/features/text_lib

Authors:
  * Heng-Jui Chang 2022
"""

import abc
import re
import tempfile
from typing import List

# Replacing the 2 tokens right before english starts as <eos> & <unk>
BERT_FIRST_IDX = 997
# Drop rest of tokens
BERT_LAST_IDX = 29635

# Default vocabularies
CHARACTER_VOCAB = list(" 'ABCDEFGHIJKLMNOPQRSTUVWXYZ")
PHONEME_VOCAB = "SIL SPN AA0 AA1 AA2 AE0 AE1 AE2 AH0 AH1 AH2 AO0 AO1 AO2 AW0 AW1 AW2 AY0 AY1 AY2 B CH D DH EH0 EH1 EH2 ER0 ER1 ER2 EY0 EY1 EY2 F G HH IH0 IH1 IH2 IY0 IY1 IY2 JH K L M N NG OW0 OW1 OW2 OY0 OY1 OY2 P R S SH T TH UH0 UH1 UH2 UW0 UW1 UW2 V W Y Z ZH".split(
    " "
)


__all__ = [
    "CharacterTokenizer",
    "CharacterSlotTokenizer",
    "SubwordTokenizer",
    "SubwordSlotTokenizer",
    "WordTokenizer",
    "PhonemeTokenizer",
    "load_tokenizer",
    "default_phoneme_tokenizer",
]


class Tokenizer:
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def encode(self, text: str, iob: str = None) -> List[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        raise NotImplementedError

    def __len__(self):
        return self.vocab_size

    @abc.abstractproperty
    def vocab_size(self) -> int:
        raise NotImplementedError

    @abc.abstractproperty
    def token_type(self) -> str:
        raise NotImplementedError

    @abc.abstractclassmethod
    def load_from_file(cls, vocab_file: str):
        raise NotImplementedError

    @property
    def pad_idx(self) -> int:
        return 0

    @property
    def eos_idx(self) -> int:
        return 1

    @property
    def unk_idx(self) -> int:
        return 2

    def __repr__(self) -> str:
        return "<{} vocab_size={}>".format(type(self).__name__, self.vocab_size)


class CharacterTokenizer(Tokenizer):
    """Character tokenizer."""

    def __init__(self, vocab_list: List[str] = None):
        super().__init__()

        if vocab_list is None:
            vocab_list = CHARACTER_VOCAB

        for tok in ["<pad>", "<eos>", "<unk>"]:
            # Note that vocab_list must not contain <pad>, <eos> and <unk>
            assert tok not in vocab_list

        # <pad> = 0, <eos> = 1, <unk> = 2
        self._vocab_list = ["<pad>", "<eos>", "<unk>"] + vocab_list
        self._vocab2idx = {v: idx for idx, v in enumerate(self._vocab_list)}

    def encode(self, s: str) -> List[int]:
        # Always strip trailing space, \r and \n
        s = s.strip("\r\n ")
        # Manually append eos to the end
        return [self.vocab_to_idx(v) for v in s] + [self.eos_idx]

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        vocabs = []
        for t, idx in enumerate(idxs):
            v = self.idx_to_vocab(idx)
            if idx == self.pad_idx or (ignore_repeat and t > 0 and idx == idxs[t - 1]):
                continue
            elif idx == self.eos_idx:
                break
            else:
                vocabs.append(v)
        return "".join(vocabs)

    @classmethod
    def load_from_file(cls, vocab_file: str = None, vocab_list: List[str] = None):
        if vocab_file is not None:
            with open(vocab_file, "r") as f:
                # Do not strip space because character based text encoder should
                # have a space token
                vocab_list = [line.strip("\r\n") for line in f]
        elif vocab_list is not None:
            pass
        else:
            raise ValueError(
                "No vocabulary information give, please specify either vocab_file or vocab_list."
            )

        return cls(vocab_list)

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_list)

    @property
    def token_type(self) -> str:
        return "character"

    def vocab_to_idx(self, vocab):
        return self._vocab2idx.get(vocab, self.unk_idx)

    def idx_to_vocab(self, idx):
        return self._vocab_list[idx]


class CharacterSlotTokenizer(Tokenizer):
    """Character tokenizer with slots."""

    def __init__(self, vocab_list: List[str], slots: List[str]):
        super().__init__()

        for tok in ["<pad>", "<eos>", "<unk>"]:
            # Note that vocab_list must not contain <pad>, <eos> and <unk>
            assert tok not in vocab_list

        # <pad> = 0, <eos> = 1, <unk> = 2
        self._vocab_list = ["<pad>", "<eos>", "<unk>"] + vocab_list
        self._vocab2idx = {v: idx for idx, v in enumerate(self._vocab_list)}
        self.space_idx = self.vocab_to_idx(" ")
        self.slots = slots
        self.slot2id = {
            self.slots[i]: (i + len(self._vocab_list)) for i in range(len(self.slots))
        }
        self.id2slot = {
            (i + len(self._vocab_list)): self.slots[i] for i in range(len(self.slots))
        }

    def encode(self, sent: str, iobs: str) -> List[int]:
        # Always strip trailing space, \r and \n
        sent = sent.strip("\r\n ")
        iobs = iobs.strip("\r\n ")
        sent = re.sub(" +", " ", sent).strip(" ")
        sent = sent.split(" ")
        iobs = iobs.split(" ")
        assert len(sent) == len(
            iobs
        ), f"transcription and iobs should have same number of words (split by space)"

        if sent[0] == "BOS":
            sent = sent[1:]
            iobs = iobs[1:]

        if sent[-1] == "EOS":
            sent = sent[:-1]
            iobs = iobs[:-1]

        tokens = []
        for i, (wrd, iob) in enumerate(zip(sent, iobs)):
            if iob != "O" and (i == 0 or iobs[i - 1] != iob):
                tokens.append(self.slot2id["B-" + iob])
            tokens += [self.vocab_to_idx(v) for v in wrd]
            if iob != "O" and (i == len(sent) - 1 or iobs[i + 1] != iob):
                tokens.append(self.slot2id["E-" + iob])
            if i == (len(sent) - 1):
                tokens.append(self.eos_idx)
            else:
                if len(tokens) > 0 and tokens[-1] != self.space_idx:
                    tokens.append(self.space_idx)
        assert tokens[-1] == self.eos_idx
        return tokens

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        vocabs = []
        for t, idx in enumerate(idxs):
            v = self.idx_to_vocab(idx)
            if idx == self.pad_idx or (ignore_repeat and t > 0 and idx == idxs[t - 1]):
                continue
            elif idx == self.eos_idx:
                break
            else:
                vocabs.append(v)
        return "".join(vocabs)

    @classmethod
    def load_from_file(cls, vocab_file: str, slots_file: str):
        with open(vocab_file, "r") as f:
            # Do not strip space because character based text encoder should
            # have a space token
            vocab_list = [line.strip("\r\n") for line in f]
        org_slots = open(slots_file).read().split("\n")
        slots = []
        for slot in [slot for slot in org_slots if slot != "O"]:
            slots.append("B-" + slot)
            slots.append("E-" + slot)
        return cls(vocab_list, slots)

    @property
    def vocab_size(self) -> int:
        return len(self._vocab_list) + len(self.slots)

    @property
    def token_type(self) -> str:
        return "character-slot"

    def vocab_to_idx(self, vocab):
        return self._vocab2idx.get(vocab, self.unk_idx)

    def idx_to_vocab(self, idx):
        idx = int(idx)
        if idx < len(self._vocab_list):
            return self._vocab_list[idx]
        else:
            token = self.id2slot[idx]
            if token[0] == "B":
                return token + " "
            elif token[0] == "E":
                return " " + token
            else:
                raise ValueError("id2slot get:", token)


class SubwordTokenizer(Tokenizer):
    """Subword tokenizer using sentencepiece."""

    def __init__(self, spm):
        super().__init__()
        if spm.pad_id() != 0 or spm.eos_id() != 1 or spm.unk_id() != 2:
            raise ValueError(
                "Please train sentencepiece model with following argument:\n"
                "--pad_id=0 --eos_id=1 --unk_id=2 --bos_id=-1 --model_type=bpe --eos_piece=<eos>"
            )
        self.spm = spm

    def encode(self, s: str) -> List[int]:
        tokens = self.spm.encode_as_ids(s)
        assert tokens[-1] == self.eos_idx
        return tokens

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        crop_idx = []
        for t, idx in enumerate(idxs):
            if idx == self.eos_idx:
                break
            elif idx == self.pad_idx or (
                ignore_repeat and t > 0 and idx == idxs[t - 1]
            ):
                continue
            else:
                crop_idx.append(idx)
        return self.spm.decode_ids(crop_idx)

    @classmethod
    def load_from_file(cls, filepath: str):
        import sentencepiece as splib

        spm = splib.SentencePieceProcessor()
        spm.load(filepath)
        spm.set_encode_extra_options("eos")
        return cls(spm)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.spm.set_encode_extra_options("eos")

    @property
    def vocab_size(self) -> int:
        return len(self.spm)

    @property
    def token_type(self) -> str:
        return "subword"


class SubwordSlotTokenizer(Tokenizer):
    """Subword tokenizer with slots."""

    def __init__(self, spm, slots):
        super().__init__()
        if spm.pad_id() != 0 or spm.eos_id() != 1 or spm.unk_id() != 2:
            raise ValueError(
                "Please train sentencepiece model with following argument:\n"
                "--pad_id=0 --eos_id=1 --unk_id=2 --bos_id=-1 --model_type=bpe --eos_piece=<eos>"
            )
        self.spm = spm
        self.slots = slots
        self.slot2id = {
            self.slots[i]: (i + len(self.spm)) for i in range(len(self.slots))
        }
        self.id2slot = {
            (i + len(self.spm)): self.slots[i] for i in range(len(self.slots))
        }

    def encode(self, sent: str, iobs: str) -> List[int]:
        # Always strip trailing space, \r and \n
        sent = sent.strip("\r\n ")
        iobs = iobs.strip("\r\n ")
        sent = re.sub(" +", " ", sent).strip(" ")
        sent = sent.split(" ")
        iobs = iobs.split(" ")
        assert len(sent) == len(
            iobs
        ), f"transcription and iobs should have same number of words (split by space)"

        if sent[0] == "BOS":
            sent = sent[1:]
            iobs = iobs[1:]

        if sent[-1] == "EOS":
            sent = sent[:-1]
            iobs = iobs[:-1]

        tokens = []
        for i, (wrd, iob) in enumerate(zip(sent, iobs)):
            if iob != "O" and (i == 0 or iobs[i - 1] != iob):
                tokens.append(self.slot2id["B-" + iob])
            encoded = self.spm.encode_as_ids(wrd)
            assert encoded[-1] == self.eos_idx
            tokens += encoded[:-1]  # drop eos
            if iob != "O" and (i == len(sent) - 1 or iobs[i + 1] != iob):
                tokens.append(self.slot2id["E-" + iob])
        assert tokens[-1] != self.eos_idx
        tokens.append(self.eos_idx)
        return tokens

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        crop_idx = []
        for t, idx in enumerate(idxs):
            if idx == self.eos_idx:
                break
            elif idx == self.pad_idx or (
                ignore_repeat and t > 0 and idx == idxs[t - 1]
            ):
                continue
            else:
                crop_idx.append(idx)

        sent, ret = [], []
        for i, x in enumerate(crop_idx):
            if x >= len(self.spm):  # x is slot token
                slot = self.id2slot[x]
                ret.append(slot)
                if len(sent) > 0:
                    decoded = self.spm.decode_ids(sent)
                    ret.insert(-1, decoded)
                    sent = []
            else:  # x is a regular token interpretable by spm
                sent.append(x)
        return " ".join(ret)

    @classmethod
    def load_from_file(cls, filepath: str, slots_file: str):
        import sentencepiece as splib

        spm = splib.SentencePieceProcessor()
        spm.load(filepath)
        spm.set_encode_extra_options(":eos")
        org_slots = open(slots_file).read().split("\n")
        slots = []
        for slot in [slot for slot in org_slots if slot != "O"]:
            slots.append("B-" + slot)
            slots.append("E-" + slot)
        return cls(spm, slots)

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.spm.set_encode_extra_options("eos")

    @property
    def vocab_size(self) -> int:
        return len(self.spm) + len(self.slots)

    @property
    def token_type(self) -> str:
        return "subword-slot"


class WordTokenizer(CharacterTokenizer):
    """Word tokenizer."""

    def encode(self, s: str) -> List[int]:
        # Always strip trailing space, \r and \n
        s = s.strip("\r\n ")
        # Space as the delimiter between words
        words = s.split(" ")
        # Manually append eos to the end
        return [self.vocab_to_idx(v) for v in words] + [self.eos_idx]

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        vocabs = []
        for t, idx in enumerate(idxs):
            v = self.idx_to_vocab(idx)
            if idx == self.eos_idx:
                break
            elif idx == self.pad_idx or (
                ignore_repeat and t > 0 and idx == idxs[t - 1]
            ):
                continue
            else:
                vocabs.append(v)
        return " ".join(vocabs)

    @property
    def token_type(self) -> str:
        return "word"


class PhonemeTokenizer(WordTokenizer):
    """Phoneme tokenizer."""

    @property
    def token_type(self) -> str:
        return "phoneme"


class BertTokenizer(Tokenizer):
    """Bert Tokenizer.

    https://github.com/huggingface/pytorch-transformers/blob/master/pytorch_transformers/tokenization_bert.py
    """

    def __init__(self, tokenizer):
        super().__init__()
        self._tokenizer = tokenizer
        self._tokenizer.pad_token = "<pad>"
        self._tokenizer.eos_token = "<eos>"
        self._tokenizer.unk_token = "<unk>"

    def encode(self, s: str) -> List[int]:
        # Reduce vocab size manually
        reduced_idx = []
        for idx in self._tokenizer.encode(s):
            try:
                r_idx = idx - BERT_FIRST_IDX
                assert r_idx > 0
                reduced_idx.append(r_idx)
            except AssertionError:
                reduced_idx.append(self.unk_idx)
        reduced_idx.append(self.eos_idx)
        return reduced_idx

    def decode(self, idxs: List[int], ignore_repeat: bool = False) -> str:
        crop_idx = []
        for t, idx in enumerate(idxs):
            if idx == self.eos_idx:
                break
            elif idx == self.pad_idx or (
                ignore_repeat and t > 0 and idx == idxs[t - 1]
            ):
                continue
            else:
                # Shift to correct idx for bert tokenizer
                crop_idx.append(idx + BERT_FIRST_IDX)
        return self._tokenizer.decode(crop_idx)

    @classmethod
    def load_from_file(cls, vocab_file: str):
        from pytorch_transformers import BertTokenizer as bert_tokenizer

        return cls(bert_tokenizer.from_pretrained(vocab_file))

    @property
    def vocab_size(self) -> int:
        return BERT_LAST_IDX - BERT_FIRST_IDX + 1

    @property
    def token_type(self) -> str:
        return "bert"


def load_tokenizer(
    mode: str,
    vocab_file: str = None,
    vocab_list: List[str] = None,
    slots_file: str = None,
) -> Tokenizer:
    """Load a text tokenizer.

    Args:
        mode (str): Mode ("character", "character-slot", "subword", "subword-slot", "word", "bert-...")
        vocab_file (str, optional): Path to vocabularies. Defaults to None.
        vocab_list (List[str], optional): List of vocabularies. Defaults to None.
        slots_file (str, optional): Path to slots. Defaults to None.

    Raises:
        NotImplementedError: If mode is not implemented.

    Returns:
        Tokenizer: Text tokenizer.
    """
    assert (
        int(vocab_file is not None) + int(vocab_list is not None) <= 1
    ), "For 'vocab_file' and 'vocab_list', at most one argument can be presented"

    with tempfile.NamedTemporaryFile("w") as f:
        if vocab_list is not None:
            f.writelines([f"{vocab}\n" for vocab in vocab_list])
            f.flush()
            vocab_file = f.name

        if slots_file is not None and not mode.endswith("slot"):
            mode = f"{mode}-slot"

        if mode == "character":
            return CharacterTokenizer.load_from_file(vocab_file)
        elif mode == "character-slot":
            return CharacterSlotTokenizer.load_from_file(vocab_file, slots_file)
        elif mode == "subword":
            return SubwordTokenizer.load_from_file(vocab_file)
        elif mode == "subword-slot":
            return SubwordSlotTokenizer.load_from_file(vocab_file, slots_file)
        elif mode == "word":
            return WordTokenizer.load_from_file(vocab_file)
        elif mode == "phoneme":
            return PhonemeTokenizer.load_from_file(vocab_file)
        elif mode.startswith("bert-"):
            return BertTokenizer.load_from_file(mode)
        else:
            raise NotImplementedError("`{}` is not yet supported.".format(mode))


def default_phoneme_tokenizer() -> PhonemeTokenizer:
    """Returns a default LibriSpeech phoneme tokenizer.

    Returns:
        PhonemeTokenizer: Vocabs include 71 phonemes
    """

    return PhonemeTokenizer.load_from_file(vocab_list=PHONEME_VOCAB)



import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.fluent_speech_commands import FluentSpeechCommands


@pytest.mark.corpus
def test_fluent_commands():
    config = dotenv_values()
    dataset_root = config["FluentSpeechCommands"]
    dataset = FluentSpeechCommands(dataset_root)
    dataset.data_split_ids
    dataset.data_split
    dataset.all_data



from omegaconf import MISSING

from .hear_esc50 import HearESC50

BEIJING_OPERA_NUM_FOLDS = 5

__all__ = ["HearBeijingOpera"]


class HearBeijingOpera(HearESC50):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
                test_fold=MISSING,
                num_folds=BEIJING_OPERA_NUM_FOLDS,
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=32,
                    shuffle=True,
                ),
                valid=dict(
                    batch_size=1,
                ),
                test=dict(
                    batch_size=1,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_layers=2,
                pooling_type="MeanPooling",
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                prediction_type="multiclass",
                scores=["top1_acc", "d_prime", "aucroc", "mAP"],
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-3,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=150000,
                log_step=100,
                eval_step=1000,
                save_step=100,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="top1_acc",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )



import torch.nn as nn

from .utils import NetworkCommonMixIn


class AudioNTT2020Task6(nn.Module, NetworkCommonMixIn):
    """DCASE2020 Task6 NTT Solution Audio Embedding Network."""

    def __init__(self, n_mels, d):
        super().__init__()

        self.features = nn.Sequential(
            nn.Conv2d(1, 64, 3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(64, 64, 3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
            nn.Conv2d(64, 64, 3, stride=1, padding=1),
            nn.BatchNorm2d(64),
            nn.ReLU(),
            nn.MaxPool2d(2, stride=2),
        )
        self.fc = nn.Sequential(
            nn.Linear(64 * (n_mels // (2**3)), d),
            nn.ReLU(),
            nn.Dropout(p=0.3),
            nn.Linear(d, d),
            nn.ReLU(),
        )
        self.d = d

    def forward(self, x):
        x = self.features(x)  # (batch, ch, mel, time)
        x = x.permute(0, 3, 2, 1)  # (batch, time, mel, ch)
        B, T, D, C = x.shape
        x = x.reshape((B, T, C * D))  # (batch, time, mel*ch)
        x = self.fc(x)
        return x


class AudioNTT2020(AudioNTT2020Task6):
    """BYOL-A General Purpose Representation Network.

    This is an extension of the DCASE 2020 Task 6 NTT Solution Audio Embedding Network.
    """

    # sample rate and embedding sizes are required model attributes for the HEAR API
    sample_rate = 16000
    embedding_size = 2048
    scene_embedding_size = embedding_size
    timestamp_embedding_size = embedding_size

    def __init__(self, n_mels=64, d=512):
        super().__init__(n_mels=n_mels, d=d)

    def forward(self, x):
        x = super().forward(x)
        x = x.mean(1) + x.amax(1)
        assert x.shape[1] == self.d and x.ndim == 2
        return x



"""
Utterance Classification Tasks

Authors
  * Leo 2022
"""

import logging
from typing import List

import numpy as np
import torch
import torch.nn.functional as F

from s3prl.dataio.encoder.category import CategoryEncoder, CategoryEncoders
from s3prl.metric import accuracy

from . import Task

logger = logging.getLogger(__name__)

__all__ = [
    "UtteranceClassifierExample",
    "UtteranceClassificationTask",
]


class UtteranceClassifierExample(torch.nn.Module):
    """
    Attributes:
        input_size: int
        output_size: int
    """

    def __init__(self, input_size=3, output_size=4):
        super().__init__()
        self._input_size = input_size
        self._output_size = output_size

    @property
    def input_size(self):
        return self._input_size

    @property
    def output_size(self):
        return self._output_size

    def forward(self, x, x_len):
        """
        Args:
            x (torch.Tensor): (batch_size, timestemps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            output (torch.Tensor): (batch_size, output_size)
        """
        assert x.size(-1) == self.input_size
        output = torch.randn(x.size(0), self.output_size)
        assert output


class UtteranceClassificationTask(Task):
    """
    Attributes:
        input_size (int): defined by model.input_size
        output_size (int): defined by len(categories)
    """

    def __init__(self, model: UtteranceClassifierExample, category: CategoryEncoder):
        """
        model.output_size should match len(categories)

        Args:
            model (UtteranceClassifier)
            category:
                encode: str -> int
                decode: int -> str
                __len__: -> int
        """

        super().__init__()
        self.model = model
        self.category = category
        assert self.model.output_size == len(category)

    def predict(self, x: torch.Tensor, x_len: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, timestamps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            logits (torch.Tensor): (batch_size, output_size)
            prediction (list): prediction strings
        """
        logits: torch.Tensor = self.model(x, x_len)
        predictions = [
            self.category.decode(index)
            for index in logits.argmax(dim=-1).detach().cpu().tolist()
        ]
        return logits, predictions

    def forward(
        self,
        _mode: str,
        x: torch.Tensor,
        x_len: torch.LongTensor,
        class_id: torch.LongTensor,
        label: List[str],
        unique_name: List[str],
        _dump_dir: str = None,
    ):
        logits, prediction = self.predict(x, x_len)
        loss = F.cross_entropy(logits, class_id)

        cacheable = dict(
            loss=loss.detach().cpu(),
            prediction=prediction,
            label=[self.category.decode(idx) for idx in class_id],
            unique_name=unique_name,
        )

        return loss, cacheable

    def reduction(self, _mode: str, cached_results: List[dict], _dump_dir: str = None):
        results = self.parse_cached_results(cached_results)
        predictions = results["prediction"]
        labels = results["label"]
        losses = results["loss"]

        acc = accuracy(predictions, labels)
        loss = (sum(losses) / len(losses)).item()

        return dict(
            loss=loss,
            accuracy=acc,
        )


class UtteranceMultiClassClassificationTask(Task):
    def __init__(self, model: UtteranceClassifierExample, categories: CategoryEncoders):
        super().__init__()
        self.model = model
        self.categories = categories
        assert self.model.output_size == len(categories)

    def predict(self, x: torch.Tensor, x_len: torch.LongTensor):
        """
        Args:
            x (torch.Tensor): (batch_size, timestamps, input_size)
            x_len (torch.LongTensor): (batch_size, )

        Return:
            logit (torch.Tensor): List[(batch_size, sub_output_size)]
            prediction (np.array): (batch_size, num_category)
        """
        logits: torch.Tensor = self.model(x, x_len)

        logit_start = 0
        sub_logits, sub_predictions = [], []
        for category in self.categories:
            logit_end = logit_start + len(category)
            sub_logit = logits[:, logit_start:logit_end]
            sub_logits.append(sub_logit)
            sub_predictions.append(
                [
                    category.decode(index)
                    for index in sub_logit.argmax(dim=-1).detach().cpu().tolist()
                ]
            )
            logit_start = logit_end
        prediction = np.array(sub_predictions, dtype="object").T

        return sub_logits, prediction

    def forward(
        self,
        _mode: str,
        x: torch.Tensor,
        x_len: torch.LongTensor,
        class_ids: torch.LongTensor,
        labels: np.ndarray,
        unique_name: List[str],
        _dump_dir: str = None,
    ):
        """
        Args:
            x: torch.Tensor, (batch_size, timestamps, input_size)
            x_len: torch.LongTensor, (batch_size)
            class_ids: torch.LongTensor, (batch_size, num_category)
            labels: np.ndarray, (batch_size, num_category)

        Return:
            loss: torch.Tensor
            prediction: np.ndarray
            label: np.ndarray
        """
        logit, prediction = self.predict(x, x_len)
        loss = sum(
            [
                F.cross_entropy(sub_logit, class_id)
                for sub_logit, class_id in zip(logit, class_ids.T)
            ]
        )

        cacheable = dict(
            loss=loss.detach().cpu(),
            prediction=prediction.tolist(),
            label=labels.tolist(),
            unique_name=unique_name,
        )

        return loss, cacheable

    def reduction(self, _mode: str, cached_results: List[dict], _dump_dir: str = None):
        results = self.parse_cached_results(cached_results)
        losses = results["loss"]
        predictions = results["prediction"]
        labels = results["label"]

        acc = accuracy(predictions, labels)
        loss = (sum(losses) / len(losses)).item()

        return dict(
            loss=loss,
            accuracy=acc,
        )



from pathlib import Path

import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.quesst14 import Quesst14


@pytest.mark.corpus
def test_quesst14_for_qbe():
    def quesst14_for_qbe(dataset_root: str):
        corpus = Quesst14(dataset_root)

        def path_to_dict(path: str):
            return dict(
                wav_path=path,
            )

        return dict(
            all_data={
                Path(path).stem: path_to_dict(path)
                for path in (corpus.valid_queries + corpus.test_queries + corpus.docs)
            },
            valid_keys=[Path(path).stem for path in corpus.valid_queries],
            test_keys=[Path(path).stem for path in corpus.test_queries],
            doc_keys=[Path(path).stem for path in corpus.docs],
        )

    quesst_root = dotenv_values()["Quesst14"]
    all_data, valid_keys, test_keys, doc_keys = quesst14_for_qbe(quesst_root).values()
    assert len(all_data) == 2714
    assert len(valid_keys) + len(test_keys) + len(doc_keys) == 2714



from pathlib import Path

from s3prl.util.audio_info import get_audio_info
from s3prl.util.pseudo_data import pseudo_audio


def test_audio_info():
    with pseudo_audio([3.0, 4.1, 1.1]) as (paths, num_samples):
        infos = get_audio_info(paths, [Path(path).stem for path in paths])

    assert infos[0]["num_frames"] == 3 * 16000



# Copyright (c) Facebook, Inc. All Rights Reserved

# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/multires_hubert/expert.py ]
#   Synopsis     [ the Multiresolution HuBERT wrapper ]
#   Author       [ Jiatong Shi ]
"""*********************************************************************************************"""

import logging
from pathlib import Path

import numpy as np
import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence

from ..interfaces import UpstreamBase
from .convert import load_converted_model

SAMPLE_RATE = 16000
EXAMPLE_SEC = 5

logger = logging.getLogger(__name__)


def upsample(x, upsample_rate):
    return x.repeat_interleave(upsample_rate, dim=1)


class UpstreamExpert(UpstreamBase):
    def __init__(self, ckpt, **kwargs):
        super().__init__(**kwargs)
        model, task_cfg = load_converted_model(ckpt)
        self.model = model
        self.task_cfg = task_cfg
        self.final_factors = []

        self.num_res = self.model.label_nums
        logger.info("num res: {}".format(self.num_res))

        # decide ratios
        feature_ds_rates = self.model.feature_ds_rates
        lcm = np.lcm.reduce(feature_ds_rates)
        self.upsample_factor = [lcm // res for res in feature_ds_rates][::-1]
        self.reverse_upsample_factor = self.upsample_factor[::-1][1:]
        logger.info("upsample_factor: {}".format(self.upsample_factor))

        if len(self.hooks) == 0:
            # Process Encoders
            for res_index in range(self.num_res - 1):
                module_name = "self.model.encoders[{}].layers".format(res_index)
                for module_id in range(len(eval(module_name))):
                    self.add_hook(
                        f"{module_name}[{module_id}]",
                        lambda input, output: input[0].transpose(0, 1),
                    )
                    self.final_factors.append(self.upsample_factor[res_index])
                self.add_hook(
                    "self.model.encoders[{}]".format(res_index),
                    lambda input, output: output[0],
                )
                self.final_factors.append(self.upsample_factor[res_index])

            # Process middle encoders
            module_name = "self.model.middle_encoder.layers"
            for module_id in range(len(eval(module_name))):
                self.add_hook(
                    f"{module_name}[{module_id}]",
                    lambda input, output: input[0].transpose(0, 1),
                )
                self.final_factors.append(self.upsample_factor[self.num_res - 1])
            self.add_hook("self.model.middle_encoder", lambda input, output: output[0])
            self.final_factors.append(self.upsample_factor[self.num_res - 1])

            # Process decoders
            for res_index in range(self.num_res - 1):
                module_name = "self.model.decoders[{}].layers".format(res_index)
                for module_id in range(len(eval(module_name))):
                    self.add_hook(
                        f"{module_name}[{module_id}]",
                        lambda input, output: input[0].transpose(0, 1),
                    )
                    self.final_factors.append(self.reverse_upsample_factor[res_index])
                self.add_hook(
                    "self.model.decoders[{}]".format(res_index),
                    lambda input, output: output[0],
                )
                self.final_factors.append(self.reverse_upsample_factor[res_index])

            def postprocess(xs):
                names, hiddens = zip(*xs)
                assert len(hiddens) == len(self.final_factors)
                hiddens = [
                    upsample(hiddens[i], self.final_factors[i])
                    for i in range(len(hiddens))
                ]
                unpad_len = min([hidden.size(1) for hidden in hiddens])
                hiddens = [hidden[:, :unpad_len, :] for hidden in hiddens]
                return list(zip(names, hiddens))

            self.hook_postprocess = postprocess

    def get_downsample_rates(self, key: str) -> int:
        return 320

    def forward(self, wavs):
        if self.task_cfg.normalize:
            wavs = [F.layer_norm(wav, wav.shape) for wav in wavs]

        device = wavs[0].device
        wav_lengths = torch.LongTensor([len(wav) for wav in wavs]).to(device)
        wav_padding_mask = ~torch.lt(
            torch.arange(max(wav_lengths)).unsqueeze(0).to(device),
            wav_lengths.unsqueeze(1),
        )
        padded_wav = pad_sequence(wavs, batch_first=True)

        features, feat_padding_mask = self.model.extract_features(
            padded_wav,
            padding_mask=wav_padding_mask,
            mask=None,
        )

        # This forward function only does the model forward
        # The return dict is then handled by UpstreamBase's hooks



from torch.utils import data


class Dataset(data.Dataset):
    def __len__(self) -> int:
        raise NotImplementedError

    def __getitem__(self, index: int):
        raise NotImplementedError

    def getinfo(self, index: int):
        raise NotImplementedError



"""
The most common and simple train/valid/test recipes
"""

from .example import CommonExample
from .hear_beijing_opera import HearBeijingOpera
from .hear_cremad import HearCremaD
from .hear_dcase_2016_task2 import HearDcase2016Task2
from .hear_esc50 import HearESC50
from .hear_fsd import HearFSD
from .hear_gsc5hr import HearGSC5hr
from .hear_gtzan import HearGtzan
from .hear_gtzan_music_speech import HearGtzanMusicSpeech
from .hear_gunshot import HearGunshot
from .hear_libricount import HearLibriCount
from .hear_maestro import HearMaestro
from .hear_nsynth5hr import HearNsynth5hr
from .hear_stroke import HearStroke
from .hear_tonic import HearTonic
from .hear_vocal import HearVocal
from .hear_vox_lingual import HearVoxLingual
from .superb_er import SuperbER
from .superb_ic import SuperbIC
from .superb_ks import SuperbKS
from .superb_sid import SuperbSID

__all__ = [
    "SuperbER",
    "SuperbIC",
    "SuperbKS",
    "SuperbSID",
    "HearFSD",
    "HearESC50",
    "HearBeijingOpera",
    "HearCremaD",
    "HearGSC5hr",
    "HearGtzanMusicSpeech",
    "HearGtzan",
    "HearGunshot",
    "HearLibriCount",
    "HearNsynth5hr",
    "HearStroke",
    "HearTonic",
    "HearVocal",
    "HearVoxLingual",
    "HearDcase2016Task2",
    "HearMaestro",
    "CommonExample",
]



# Copyright Hear Benchmark Team
# Copyright Shu-wen Yang (refactor from https://github.com/hearbenchmark/hear-eval-kit)

import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import more_itertools
import numpy as np
import pandas as pd
import torch
from scipy.ndimage import median_filter
from sklearn.model_selection import ParameterGrid
from tqdm import tqdm

from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.task.base import Task

from ._hear_score import available_scores, validate_score_return_type

logger = logging.getLogger(__name__)

SAMPLE_RATE = 16000

__all__ = ["EventPredictionTask"]


def create_events_from_prediction(
    prediction_dict: Dict[float, torch.Tensor],
    idx_to_label: Dict[int, str],
    threshold: float = 0.5,
    median_filter_ms: float = 150,
    min_duration: float = 60.0,
) -> List[Dict[str, Union[float, str]]]:
    """
    Takes a set of prediction tensors keyed on timestamps and generates events.
    (This is for one particular audio scene.)
    We convert the prediction tensor to a binary label based on the threshold value. Any
    events occurring at adjacent timestamps are considered to be part of the same event.
    This loops through and creates events for each label class.
    We optionally apply median filtering to predictions.
    We disregard events that are less than the min_duration milliseconds.

    Args:
        prediction_dict: A dictionary of predictions keyed on timestamp
            {timestamp -> prediction}. The prediction is a tensor of label
            probabilities.
        idx_to_label: Index to label mapping.
        threshold: Threshold for determining whether to apply a label
        min_duration: the minimum duration in milliseconds for an
                event to be included.

    Returns:
        A list of dicts withs keys "label", "start", and "end"
    """
    # Make sure the timestamps are in the correct order
    timestamps = np.array(sorted(prediction_dict.keys()))

    # Create a sorted numpy matrix of frame level predictions for this file. We convert
    # to a numpy array here before applying a median filter.
    predictions = np.stack(
        [prediction_dict[t].detach().cpu().numpy() for t in timestamps]
    )

    # Optionally apply a median filter here to smooth out events.
    ts_diff = np.mean(np.diff(timestamps))
    if median_filter_ms:
        filter_width = int(round(median_filter_ms / ts_diff))
        if filter_width:
            predictions = median_filter(predictions, size=(filter_width, 1))

    # Convert probabilities to binary vectors based on threshold
    predictions = (predictions > threshold).astype(np.int8)

    events = []
    for label in range(predictions.shape[1]):
        for group in more_itertools.consecutive_groups(
            np.where(predictions[:, label])[0]
        ):
            grouptuple = tuple(group)
            assert (
                tuple(sorted(grouptuple)) == grouptuple
            ), f"{sorted(grouptuple)} != {grouptuple}"
            startidx, endidx = (grouptuple[0], grouptuple[-1])

            start = timestamps[startidx]
            end = timestamps[endidx]
            # Add event if greater than the minimum duration threshold
            if end - start >= min_duration:
                events.append(
                    {"label": idx_to_label[label], "start": start, "end": end}
                )

    # This is just for pretty output, not really necessary
    events.sort(key=lambda k: k["start"])
    return events


def get_events_for_all_files(
    predictions: torch.Tensor,
    filenames: List[str],
    timestamps: torch.Tensor,
    idx_to_label: Dict[int, str],
    postprocessing_grid: Dict[str, List[float]],
    postprocessing: Optional[Tuple[Tuple[str, Any], ...]] = None,
) -> Dict[Tuple[Tuple[str, Any], ...], Dict[str, List[Dict[str, Union[str, float]]]]]:
    """
    Produces lists of events from a set of frame based label probabilities.
    The input prediction tensor may contain frame predictions from a set of different
    files concatenated together. file_timestamps has a list of filenames and
    timestamps for each frame in the predictions tensor.

    We split the predictions into separate tensors based on the filename and compute
    events based on those individually.

    If no postprocessing is specified (during training), we try a
    variety of ways of postprocessing the predictions into events,
    from the postprocessing_grid including median filtering and
    minimum event length.

    If postprocessing is specified (during test, chosen at the best
    validation epoch), we use this postprocessing.

    Args:
        predictions: a tensor of frame based multi-label predictions.
        filenames: a list of filenames where each entry corresponds
            to a frame in the predictions tensor.
        timestamps: a list of timestamps where each entry corresponds
            to a frame in the predictions tensor.
        idx_to_label: Index to label mapping.
        postprocessing: See above.

    Returns:
        A dictionary from filtering params to the following values:
        A dictionary of lists of events keyed on the filename slug.
        The event list is of dicts of the following format:
            {"label": str, "start": float ms, "end": float ms}
    """
    # This probably could be more efficient if we make the assumption that
    # timestamps are in sorted order. But this makes sure of it.
    assert predictions.shape[0] == len(filenames)
    assert predictions.shape[0] == len(timestamps)
    event_files: Dict[str, Dict[float, torch.Tensor]] = {}
    for i, (filename, timestamp) in enumerate(zip(filenames, timestamps)):
        slug = Path(filename).name

        # Key on the slug to be consistent with the ground truth
        if slug not in event_files:
            event_files[slug] = {}

        # Save the predictions for the file keyed on the timestamp
        event_files[slug][float(timestamp) * 1000] = predictions[i]

    # Create events for all the different files. Store all the events as a dictionary
    # with the same format as the ground truth from the luigi pipeline.
    # Ex) { slug -> [{"label" : "woof", "start": 0.0, "end": 2.32}, ...], ...}
    event_dict: Dict[
        Tuple[Tuple[str, Any], ...], Dict[str, List[Dict[str, Union[float, str]]]]
    ] = {}
    if postprocessing:
        postprocess = postprocessing
        event_dict[postprocess] = {}
        logger.info("Use searched postprocess config to decode")
        for slug, timestamp_predictions in event_files.items():
            event_dict[postprocess][slug] = create_events_from_prediction(
                timestamp_predictions, idx_to_label, **dict(postprocess)
            )
    else:
        postprocessing_confs = list(ParameterGrid(postprocessing_grid))
        for postprocess_dict in tqdm(
            postprocessing_confs, desc="Search postprocessing"
        ):
            postprocess = tuple(postprocess_dict.items())
            event_dict[postprocess] = {}
            for slug, timestamp_predictions in event_files.items():
                event_dict[postprocess][slug] = create_events_from_prediction(
                    timestamp_predictions, idx_to_label, **postprocess_dict
                )

    return event_dict


def label_vocab_nlabels(embedding_path: Path) -> Tuple[pd.DataFrame, int]:
    label_vocab = pd.read_csv(embedding_path.joinpath("labelvocabulary.csv"))

    nlabels = len(label_vocab)
    assert nlabels == label_vocab["idx"].max() + 1
    return (label_vocab, nlabels)


class OneHotToCrossEntropyLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.loss = torch.nn.CrossEntropyLoss()

    def forward(self, y_hat: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        assert torch.all(
            torch.sum(y, dim=1) == torch.ones(y.shape[0], device=self.device)
        )
        y = y.argmax(dim=1)
        return self.loss(y_hat, y)


class EventPredictionTask(Task):
    def __init__(
        self,
        model: torch.nn.Module,
        category: CategoryEncoder,
        prediction_type: str,
        scores: List[str],
        postprocessing_grid: Dict[str, List[float]],
        valid_target_events: Dict[str, List[Dict[str, Any]]] = None,
        test_target_events: Dict[str, List[Dict[str, Any]]] = None,
    ):
        super().__init__()
        self.model = model
        assert isinstance(self.model.downsample_rate, int)
        self.feat_frame_shift = self.model.downsample_rate

        self.label_to_idx = {
            str(category.decode(idx)): idx for idx in range(len(category))
        }
        self.idx_to_label = {
            idx: str(category.decode(idx)) for idx in range(len(category))
        }
        self.scores = [
            available_scores[score](label_to_idx=self.label_to_idx) for score in scores
        ]

        if prediction_type == "multilabel":
            self.activation: torch.nn.Module = torch.nn.Sigmoid()
            self.logit_loss = torch.nn.BCEWithLogitsLoss()
        elif prediction_type == "multiclass":
            self.activation = torch.nn.Softmax()
            self.logit_loss = OneHotToCrossEntropyLoss()
        else:
            raise ValueError(f"Unknown prediction_type {prediction_type}")

        self.target_events = {
            "valid": valid_target_events,
            "test": test_target_events,
        }
        # For each epoch, what postprocessing parameters were best
        self.postprocessing_grid = postprocessing_grid
        self.best_postprocessing = None

    def get_state(self):
        return {
            "best_postprocessing": self.best_postprocessing,
        }

    def set_state(self, state: dict):
        self.best_postprocessing = state["best_postprocessing"]

    def predict(self, x, x_len):
        logits, _ = self.model(x, x_len)
        prediction = self.activation(logits)
        return prediction, logits, x_len

    def _match_length(self, inputs, labels):
        """
        Since the upstream extraction process can sometimes cause a mismatch
        between the seq lenth of inputs and labels:
        - if len(inputs) > len(labels), we truncate the final few timestamp of inputs to match the length of labels
        - if len(inputs) < len(labels), we duplicate the last timestep of inputs to match the length of labels
        Note that the length of labels should never be changed.
        """
        input_len, label_len = inputs.size(1), labels.size(1)

        factor = int(round(label_len / input_len))
        assert factor == 1
        if input_len > label_len:
            inputs = inputs[:, :label_len, :]
        elif input_len < label_len:
            pad_vec = inputs[:, -1, :].unsqueeze(1)  # (batch_size, 1, feature_dim)
            inputs = torch.cat(
                (inputs, pad_vec.repeat(1, label_len - input_len, 1)), dim=1
            )  # (batch_size, seq_len, feature_dim), where seq_len == labels.size(-1)
        return inputs

    def train_step(
        self,
        x,
        x_len,
        y,
        y_len,
        labels,
        record_id: List[str],
        chunk_id: List[int],
        unique_name: List[str],
        _dump_dir: str = None,
    ):
        y_hat, y_hat_len = self.model(x, x_len)
        y_hat = self._match_length(y_hat, y)

        assert y_hat.size(-1) == y.size(-1), f"{y_hat.size(-1)} == {y.size(-1)}"

        hidden_size = y_hat.size(-1)
        loss = self.logit_loss(
            y_hat.reshape(-1, hidden_size).float(), y.reshape(-1, hidden_size).float()
        )

        cacheable = {
            "loss": loss.detach().cpu().item(),
        }

        return loss, cacheable

    def train_reduction(self, batch_results: list, _dump_dir: str = None):
        loss = []
        for batch in batch_results:
            loss.append(batch["loss"])
        loss = torch.FloatTensor(loss).mean().item()

        return {
            "loss": loss,
        }

    def _eval_step(
        self,
        x,
        x_len,
        y,
        y_len,
        labels,
        record_id: List[str],
        chunk_id: List[int],
        unique_name: List[str],
        _dump_dir: str = None,
    ):
        y_pr, y_hat, y_pr_len = self.predict(x, x_len)
        y_pr = self._match_length(y_pr, y)
        y_hat = self._match_length(y_hat, y)

        assert len(set(record_id)) == 1
        chunk_id = chunk_id.detach().cpu().tolist()
        assert sorted(chunk_id) == chunk_id

        y_pr_trim, y_hat_trim, y_trim = [], [], []
        for _p, _h, _y, length in zip(y_pr, y_hat, y, y_len):
            y_pr_trim.append(_p[:length])
            y_hat_trim.append(_h[:length])
            y_trim.append(_y[:length])
        y_pr_trim = torch.cat(y_pr_trim, dim=0)
        y_hat_trim = torch.cat(y_hat_trim, dim=0)
        y_trim = torch.cat(y_trim, dim=0)

        return 0, dict(
            label=y_trim,  # (seqlen, num_class)
            logit=y_hat_trim,  # (seqlen, num_class)
            prediction=y_pr_trim,  # (seqlen, num_class)
            record_id=record_id[0],  # List[str]
        )

    def valid_step(self, *args, **kwds):
        return self._eval_step(*args, **kwds)

    def test_step(self, *args, **kwds):
        return self._eval_step(*args, **kwds)

    def log_scores(self, score_args):
        """Logs the metric score value for each score defined for the model"""
        assert hasattr(self, "scores"), "Scores for the model should be defined"
        end_scores = {}
        # The first score in the first `self.scores` is the optimization criterion
        for score in self.scores:
            score_ret = score(*score_args)
            validate_score_return_type(score_ret)
            # If the returned score is a tuple, store each subscore as separate entry
            if isinstance(score_ret, tuple):
                end_scores[f"{score}"] = score_ret[0][1]
                # All other scores will also be logged
                for subscore, value in score_ret:
                    end_scores[f"{score}_{subscore}"] = value
            elif isinstance(score_ret, float):
                end_scores[f"{score}"] = score_ret
            else:
                raise ValueError(
                    f"Return type {type(score_ret)} is unexpected. Return type of "
                    "the score function should either be a "
                    "tuple(tuple) or float."
                )
        return end_scores

    def valid_reduction(self, cached_results: list, _dump_dir: str = None):
        return self.eval_reduction("valid", cached_results, _dump_dir)

    def test_reduction(self, cached_results: list, _dump_dir: str = None):
        return self.eval_reduction("test", cached_results, _dump_dir)

    def eval_reduction(self, _mode: str, cached_results: list, _dump_dir: str = None):
        target, prediction, prediction_logit, filename, timestamp = [], [], [], [], []
        for batch in cached_results:
            length = batch["label"].size(0)
            assert batch["prediction"].size(0) == length
            assert batch["logit"].size(0) == length

            target.append(batch["label"])
            prediction.append(batch["prediction"])
            prediction_logit.append(batch["logit"])
            filename += [batch["record_id"]] * length

            ts = (
                torch.arange(1, length + 1).float() * self.feat_frame_shift
                - self.feat_frame_shift / 2
            ) / SAMPLE_RATE
            timestamp += ts.tolist()

        target = torch.cat(target, dim=0)  # (timestamp, hidden_size)
        prediction = torch.cat(prediction, dim=0)
        prediction_logit = torch.cat(prediction_logit, dim=0)
        timestamp = torch.FloatTensor(timestamp)

        loss = self.logit_loss(prediction_logit.float(), target.float())

        logs = {"loss": loss.detach().cpu().item()}

        if _mode in ["valid", "test"]:
            # events in miniseconds
            predicted_events_by_postprocessing = get_events_for_all_files(
                prediction,
                filename,
                timestamp,
                self.idx_to_label,
                self.postprocessing_grid,
                self.best_postprocessing if _mode == "test" else None,
            )

            score_and_postprocessing = []
            for postprocessing in tqdm(predicted_events_by_postprocessing):
                predicted_events = predicted_events_by_postprocessing[postprocessing]
                primary_score_fn = self.scores[0]
                primary_score_ret = primary_score_fn(
                    predicted_events, self.target_events[_mode]
                )
                if isinstance(primary_score_ret, tuple):
                    primary_score = primary_score_ret[0][1]
                elif isinstance(primary_score_ret, float):
                    primary_score = primary_score_ret
                else:
                    raise ValueError(
                        f"Return type {type(primary_score_ret)} is unexpected. "
                        "Return type of the score function should either be a "
                        "tuple(tuple) or float. "
                    )
                if np.isnan(primary_score):
                    primary_score = 0.0
                score_and_postprocessing.append((primary_score, postprocessing))
            score_and_postprocessing.sort(reverse=True)

            if _mode in ["valid", "test"]:
                self.best_postprocessing = score_and_postprocessing[0][1]
                logger.info(f"Best postprocessing: {self.best_postprocessing}")

            predicted_events = predicted_events_by_postprocessing[
                self.best_postprocessing
            ]

            if _mode == "test":
                self.test_predictions = {
                    "target": target.detach().cpu(),
                    "prediction": prediction.detach().cpu(),
                    "prediction_logit": prediction_logit.detach().cpu(),
                    "target_events": self.target_events[_mode],
                    "predicted_events": predicted_events,
                    "timestamp": timestamp,
                }

            score_logs = self.log_scores(
                score_args=(predicted_events, self.target_events[_mode])
            )
            logs.update(score_logs)

        return logs



"""
Limit the maximum timestamps in a batch to realize dynamic batching.

Authors:
  * Leo 2022
"""

from typing import List

import torch

__all__ = [
    "MaxTimestampBatchSampler",
]


class MaxTimestampBatchSampler:
    """
    The reduced timestamps for a batch should not exceed the max_timestamp.
    If shuffled, each indices are first shuffled before aggregated into batches
    """

    def __init__(
        self,
        lengths: List[int],
        max_length: int,
        shuffle: bool = False,
        seed: int = 12345678,
        reduce_func: callable = None,
    ) -> None:
        self.lengths = lengths
        self.max_length = max_length
        self.shuffle = shuffle
        self.seed = seed
        self.epoch = 0
        self.reduce_func = reduce_func or self._default_reduce_func

    @staticmethod
    def _default_reduce_func(timestamps):
        return max(timestamps) * len(timestamps)

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def _evaluate_reduced_timestamps(self, batch_indices):
        return self.reduce_func([self.lengths[indice] for indice in batch_indices])

    def __iter__(self):
        if self.shuffle:
            generator = torch.Generator()
            generator.manual_seed(self.epoch + self.seed)
            indices = torch.randperm(len(self.lengths), generator=generator).tolist()
        else:
            indices = list(range(len(self.lengths)))

        batch = []
        for indice in indices:
            try_new_batch = batch + [indice]
            if self._evaluate_reduced_timestamps(try_new_batch) <= self.max_length:
                batch = try_new_batch
            elif len(batch) == 0:
                raise ValueError(
                    f"There is a single length {self.lengths[indice]} larger than "
                    f"max_length {self.max_length}. Please increase "
                    "the max_length."
                )
            else:
                yield batch
                batch = [indice]

        if len(batch) > 0:
            yield batch

    def __len__(self):
        return len(list(iter(self)))
